from boa3.boa3 import Boa3
from boa3.neo import to_script_hash
from boa3.neo.vm.type.String import String
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.TestExecutionException import TestExecutionException
from boa3_test.tests.test_classes.testengine import TestEngine


class TestTemplate(BoaTest):
    default_folder: str = 'examples'

    OWNER_SCRIPT_HASH = bytes(20)
    OTHER_ACCOUNT_1 = to_script_hash(b'NiNmXL8FjEUEs1nfX9uHFBNaenxDHJtmuB')

    def test_nep5_compile(self):
        path = self.get_contract_path('nep5.py')
        Boa3.compile(path)

    def test_nep5_name(self):
        path = self.get_contract_path('nep5.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'name')
        self.assertEqual('NEP5 Standard', result)

    def test_nep5_symbol(self):
        path = self.get_contract_path('nep5.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'symbol')
        self.assertEqual('NEP5', result)

    def test_nep5_decimals(self):
        path = self.get_contract_path('nep5.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'decimals')
        self.assertEqual(8, result)

    def test_nep5_total_supply(self):
        total_supply = 10_000_000 * 10 ** 8

        path = self.get_contract_path('nep5.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'totalSupply')
        self.assertEqual(total_supply, result)

    def test_nep5_total_balance_of(self):
        total_supply = 10_000_000 * 10 ** 8

        path = self.get_contract_path('nep5.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'balanceOf', self.OWNER_SCRIPT_HASH)
        self.assertEqual(total_supply, result)

        # should fail when the script length is not 20
        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'balanceOf', bytes(10))
        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'balanceOf', bytes(30))

    def test_nep5_total_transfer(self):
        transferred_amount = 10 * 10 ** 8  # 10 tokens

        path = self.get_contract_path('nep5.py')
        engine = TestEngine()

        # should fail if the sender doesn't sign
        result = self.run_smart_contract(engine, path, 'transfer',
                                         self.OWNER_SCRIPT_HASH, self.OTHER_ACCOUNT_1, transferred_amount)
        self.assertEqual(False, result)

        # other account doesn't have enough balance
        result = self.run_smart_contract(engine, path, 'transfer',
                                         self.OTHER_ACCOUNT_1, self.OWNER_SCRIPT_HASH, transferred_amount,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(False, result)

        # should fail when any of the scripts' length is not 20
        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'transfer',
                                    self.OWNER_SCRIPT_HASH, bytes(10), transferred_amount)
        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'transfer',
                                    bytes(10), self.OTHER_ACCOUNT_1, transferred_amount)

        # should fail when the amount is less than 0
        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'transfer',
                                    self.OTHER_ACCOUNT_1, self.OWNER_SCRIPT_HASH, -10)

        # doesn't fire the transfer event when transferring to yourself
        balance_before = self.run_smart_contract(engine, path, 'balanceOf', self.OWNER_SCRIPT_HASH)
        result = self.run_smart_contract(engine, path, 'transfer',
                                         self.OWNER_SCRIPT_HASH, self.OWNER_SCRIPT_HASH, transferred_amount,
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(True, result)
        transfer_events = engine.get_events('transfer')
        # there is one transfer event thanks to the deploy
        self.assertEqual(1, len(transfer_events))

        # transferring to yourself doesn't change the balance
        balance_after = self.run_smart_contract(engine, path, 'balanceOf', self.OWNER_SCRIPT_HASH)
        self.assertEqual(balance_before, balance_after)

        # doesn't fire the transfer event when transferring to yourself
        balance_sender_before = self.run_smart_contract(engine, path, 'balanceOf', self.OWNER_SCRIPT_HASH)
        balance_receiver_before = self.run_smart_contract(engine, path, 'balanceOf', self.OTHER_ACCOUNT_1)
        result = self.run_smart_contract(engine, path, 'transfer',
                                         self.OWNER_SCRIPT_HASH, self.OTHER_ACCOUNT_1, transferred_amount,
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(True, result)
        transfer_events = engine.get_events('transfer')
        self.assertEqual(2, len(transfer_events))
        self.assertEqual(3, len(transfer_events[1].arguments))

        sender, receiver, amount = transfer_events[1].arguments
        if isinstance(sender, str):
            sender = String(sender).to_bytes()
        if isinstance(receiver, str):
            receiver = String(sender).to_bytes()
        self.assertEqual(self.OWNER_SCRIPT_HASH, sender)
        self.assertEqual(self.OTHER_ACCOUNT_1, receiver)
        self.assertEqual(transferred_amount, amount)

        # transferring to yourself doesn't change the balance
        balance_sender_after = self.run_smart_contract(engine, path, 'balanceOf', self.OWNER_SCRIPT_HASH)
        balance_receiver_after = self.run_smart_contract(engine, path, 'balanceOf', self.OTHER_ACCOUNT_1)
        self.assertEqual(balance_sender_before - transferred_amount, balance_sender_after)
        self.assertEqual(balance_receiver_before + transferred_amount, balance_receiver_after)

    def test_nep5_verify(self):
        path = self.get_contract_path('nep5.py')
        engine = TestEngine()

        # should fail without signature
        result = self.run_smart_contract(engine, path, 'verify')
        self.assertEqual(False, result)

        # should fail if not signed by the owner
        result = self.run_smart_contract(engine, path, 'verify',
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, path, 'verify',
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(True, result)



import ast
from typing import Any, Dict, List, Optional

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.variable import Variable


class IndexMethod(IBuiltinMethod):

    def __init__(self, args: Dict[str, Variable] = None, defaults: List[ast.AST] = None):
        from boa3.model.type.type import Type
        identifier = 'index'
        super().__init__(identifier, args, defaults=defaults, return_type=Type.int)

    @property
    def identifier(self) -> str:
        from boa3.model.type.type import Type

        if self._arg_self.type is Type.sequence:  # IndexSequenceMethod default value for self
            return self._identifier

        if self._arg_self.type is Type.str:
            return '-{0}_{1}'.format(self._identifier, Type.str.identifier)

        if Type.sequence.is_type_of(self._arg_self.type):
            return '-{0}_{1}'.format(self._identifier, Type.sequence.identifier)

        return self._identifier

    @property
    def is_supported(self) -> bool:
        # TODO: change when index() with only one argument is implemented for range
        from boa3.model.type.type import Type
        if Type.range.is_type_of(self._arg_self.type):
            return False
        return True

    @property
    def _arg_self(self) -> Variable:
        return self.args['self']

    @property
    def _arg_value(self) -> Variable:
        return self.args['value']

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return

    def build(self, value: Any) -> IBuiltinMethod:
        if not isinstance(value, list):
            value = [value]

        from boa3.model.type.collection.sequence.mutable.listtype import ListType
        from boa3.model.type.collection.sequence.tupletype import TupleType
        from boa3.model.type.collection.sequence.rangetype import RangeType
        if len(value) > 1 and isinstance(value[0], (ListType, TupleType, RangeType)):
            from boa3.model.builtin.classmethod.indexsequencemethod import IndexSequenceMethod
            return IndexSequenceMethod(value[0], value[1])

        from boa3.model.type.type import Type
        if len(value) > 0 and (Type.str.is_type_of(value[0]) or Type.bytes.is_type_of(value[0])):
            from boa3.model.builtin.classmethod.indexbytesstringmethod import IndexBytesStringMethod
            return IndexBytesStringMethod(value[0])

        return super().build(value)



from .ContractManagement import ContractManagement, ContractManagementMethod
from .CryptoLib import CryptoLibContract, CryptoLibMethod
from .Ledger import LedgerContract, LedgerMethod
from .Neo import NeoContract, NeoContractMethod
from .Nep17 import Nep17Contract, Nep17Method
from .Oracle import OracleContract, OracleMethod
from .PolicyContract import PolicyContract, PolicyContractMethod
from .RoleManagement import RoleManagement, RoleManagementMethod
from .StdLib import StdLibContract, StdLibMethod



from typing import Dict, Optional

from boa3.model.builtin.method import ByteArrayMethod
from boa3.model.variable import Variable


class ByteArrayEncodingMethod(ByteArrayMethod):

    def __init__(self):
        from boa3.model.type.type import Type

        args: Dict[str, Variable] = {
            'object': Variable(Type.str),
            'encoding': Variable(Type.str)
        }
        super().__init__(args)

    @property
    def is_supported(self) -> bool:
        # TODO: change when building bytearray from string encoding
        return False

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return



from boa3 import constants
from boa3.boa3 import Boa3
from boa3.neo import to_script_hash
from boa3.neo.vm.type.String import String
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.TestExecutionException import TestExecutionException
from boa3_test.tests.test_classes.testengine import TestEngine


class TestTemplate(BoaTest):
    default_folder: str = 'examples'

    OWNER_SCRIPT_HASH = bytes(20)
    OTHER_ACCOUNT_1 = to_script_hash(b'NiNmXL8FjEUEs1nfX9uHFBNaenxDHJtmuB')
    OTHER_ACCOUNT_2 = bytes(range(20))

    def test_amm_compile(self):
        path = self.get_contract_path('amm.py')
        Boa3.compile(path)

    def test_amm_set_address(self):
        path = self.get_contract_path('amm.py')
        path_zneo = self.get_contract_path('wrapped_neo.py')
        path_zgas = self.get_contract_path('wrapped_gas.py')
        engine = TestEngine()

        self.run_smart_contract(engine, path_zneo, 'symbol')
        zneo_address = engine.executed_script_hash.to_array()

        self.run_smart_contract(engine, path_zgas, 'symbol')
        zgas_address = engine.executed_script_hash.to_array()

        engine.reset_engine()
        engine.add_contract(path_zneo.replace('.py', '.nef'))
        engine.add_contract(path_zgas.replace('.py', '.nef'))

        # won't work because it needs the owner signature
        result = self.run_smart_contract(engine, path, 'set_address', zneo_address, zgas_address,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(False, result)

        # it will work now
        result = self.run_smart_contract(engine, path, 'set_address', zneo_address, zgas_address,
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(True, result)

        # initialize will work once
        result = self.run_smart_contract(engine, path, 'set_address', zneo_address, zgas_address,
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(False, result)

    def test_amm_symbol(self):
        path = self.get_contract_path('amm.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'symbol')
        self.assertEqual('AMM', result)

    def test_amm_decimals(self):
        path = self.get_contract_path('amm.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'decimals')
        self.assertEqual(8, result)

    def test_amm_total_supply(self):
        path = self.get_contract_path('amm.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'totalSupply')
        self.assertEqual(0, result)

    def test_amm_total_balance_of(self):
        path = self.get_contract_path('amm.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'balanceOf', self.OWNER_SCRIPT_HASH)
        self.assertEqual(0, result)

        # should fail when the script length is not 20
        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'balanceOf', bytes(10))
        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'balanceOf', bytes(30))

    def test_amm_quote(self):
        path = self.get_contract_path('amm.py')
        engine = TestEngine()
        amount_zneo = 1 * 10 ** 8
        reserve_zneo = 100 * 10 ** 8
        reserve_zgas = 1100 * 10 ** 8
        result = self.run_smart_contract(engine, path, 'quote', amount_zneo, reserve_zneo, reserve_zgas)
        amount_zgas = amount_zneo * reserve_zgas // reserve_zneo
        self.assertEqual(amount_zgas, result)

    def test_amm_on_nep17_payment(self):
        transferred_amount = 10 * 10 ** 8

        path = self.get_contract_path('amm.py')
        path_zneo = self.get_contract_path('wrapped_neo.py')
        path_zgas = self.get_contract_path('wrapped_gas.py')
        engine = TestEngine()

        self.run_smart_contract(engine, path, 'symbol')
        amm_address = engine.executed_script_hash.to_array()

        self.run_smart_contract(engine, path_zneo, 'symbol')
        zneo_address = engine.executed_script_hash.to_array()

        self.run_smart_contract(engine, path_zgas, 'symbol')
        zgas_address = engine.executed_script_hash.to_array()

        engine.reset_engine()
        engine.add_contract(path.replace('.py', '.nef'))
        engine.add_contract(path_zneo.replace('.py', '.nef'))
        engine.add_contract(path_zgas.replace('.py', '.nef'))

        result = self.run_smart_contract(engine, path, 'set_address', zneo_address, zgas_address,
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(True, result)

        # the smart contract will abort if some address other than zNEO or zGAS calls the onPayment method
        with self.assertRaisesRegex(TestExecutionException, self.ABORTED_CONTRACT_MSG):
            self.run_smart_contract(engine, path, 'onNEP17Payment', self.OTHER_ACCOUNT_1, transferred_amount, None,
                                    signer_accounts=[self.OTHER_ACCOUNT_1])

        engine.add_neo(self.OTHER_ACCOUNT_1, transferred_amount)

        # adding the transferred_amount into OTHER_ACCOUNT_1
        result = self.run_smart_contract(engine, constants.NEO_SCRIPT, 'transfer',
                                         self.OTHER_ACCOUNT_1, zneo_address, transferred_amount, None,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(True, result)

        # the AMM will accept this transaction, but there is no reason to send tokens directly to the smart contract.
        # to send tokens to the AMM you should use the add_liquidity function
        result = self.run_smart_contract(engine, path_zneo, 'transfer',
                                         self.OTHER_ACCOUNT_1, amm_address, transferred_amount, None,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(True, result)

    def test_amm_add_liquidity(self):
        transferred_amount_zneo = 10 * 10 ** 8
        transferred_amount_zgas = 110 * 10 ** 8

        path = self.get_contract_path('amm.py')
        path_zneo = self.get_contract_path('wrapped_neo.py')
        path_zgas = self.get_contract_path('wrapped_gas.py')
        engine = TestEngine()

        self.run_smart_contract(engine, path, 'symbol')
        amm_address = engine.executed_script_hash.to_array()

        self.run_smart_contract(engine, path_zneo, 'symbol')
        zneo_address = engine.executed_script_hash.to_array()

        self.run_smart_contract(engine, path_zgas, 'symbol')
        zgas_address = engine.executed_script_hash.to_array()

        engine.reset_engine()
        engine.add_contract(path.replace('.py', '.nef'))
        engine.add_contract(path_zneo.replace('.py', '.nef'))
        engine.add_contract(path_zgas.replace('.py', '.nef'))

        result = self.run_smart_contract(engine, path, 'set_address', zneo_address, zgas_address,
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(True, result)

        engine.add_neo(self.OTHER_ACCOUNT_1, 10_000_000 * 10 ** 8)
        engine.add_gas(self.OTHER_ACCOUNT_1, 10_000_000 * 10 ** 8)

        # minting zNEO to OTHER_ACCOUNT_1
        result = self.run_smart_contract(engine, constants.NEO_SCRIPT, 'transfer',
                                         self.OTHER_ACCOUNT_1, zneo_address, 10_000_000 * 10 ** 8, None,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(True, result)

        # minting zGAS to OTHER_ACCOUNT_1
        result = self.run_smart_contract(engine, constants.GAS_SCRIPT, 'transfer',
                                         self.OTHER_ACCOUNT_1, zgas_address, 10_000_000 * 10 ** 8, None,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(True, result)

        # won't work, because the user did not allow the amm to transfer zNEO and zGAS
        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'add_liquidity', transferred_amount_zneo, transferred_amount_zgas, 0,
                                    0, self.OTHER_ACCOUNT_1,
                                    calling_script_hash=self.OTHER_ACCOUNT_1,
                                    signer_accounts=[self.OTHER_ACCOUNT_1])

        # approving the AMM contract, so that it will be able to transfer zNEO from OTHER_ACCOUNT_1
        result = self.run_smart_contract(engine, path_zneo, 'approve',
                                         amm_address, transferred_amount_zneo,
                                         calling_script_hash=self.OTHER_ACCOUNT_1,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(True, result)

        # approving the AMM contract, so that it will be able to transfer zGAS from OTHER_ACCOUNT_1
        result = self.run_smart_contract(engine, path_zgas, 'approve',
                                         amm_address, transferred_amount_zgas,
                                         calling_script_hash=self.OTHER_ACCOUNT_1,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(True, result)

        # saving data to demonstrate that the value will change later
        total_supply_before = self.run_smart_contract(engine, path, 'totalSupply')
        balance_user_amm_before = self.run_smart_contract(engine, path, 'balanceOf', self.OTHER_ACCOUNT_1)
        reserves_before = self.run_smart_contract(engine, path, 'get_reserves')
        balance_user_zneo_before = self.run_smart_contract(engine, path_zneo, 'balanceOf', self.OTHER_ACCOUNT_1)
        balance_user_zgas_before = self.run_smart_contract(engine, path_zgas, 'balanceOf', self.OTHER_ACCOUNT_1)
        balance_amm_zneo_before = self.run_smart_contract(engine, path_zneo, 'balanceOf', amm_address)
        balance_amm_zgas_before = self.run_smart_contract(engine, path_zgas, 'balanceOf', amm_address)
        # adding liquidity to the pool will give you AMM tokens in return
        result = self.run_smart_contract(engine, path, 'add_liquidity',
                                         transferred_amount_zneo, transferred_amount_zgas, 0, 0, self.OTHER_ACCOUNT_1,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        import math
        liquidity = int(math.sqrt(transferred_amount_zneo * transferred_amount_zgas))
        self.assertEqual(3, len(result))
        self.assertEqual(transferred_amount_zneo, result[0])
        self.assertEqual(transferred_amount_zgas, result[1])
        self.assertEqual(liquidity, result[2])

        transfer_events = engine.get_events('Transfer', origin=amm_address)
        self.assertEqual(1, len(transfer_events))
        self.assertEqual(3, len(transfer_events[0].arguments))

        sender, receiver, amount = transfer_events[0].arguments
        if isinstance(sender, str):
            sender = String(sender).to_bytes()
        if isinstance(receiver, str):
            receiver = String(receiver).to_bytes()
        self.assertEqual(None, sender)
        self.assertEqual(self.OTHER_ACCOUNT_1, receiver)
        self.assertEqual(liquidity, amount)

        sync_events = engine.get_events('Sync', origin=amm_address)
        self.assertEqual(1, len(sync_events))
        self.assertEqual(2, len(sync_events[0].arguments))

        balance_zneo, balance_zgas = sync_events[0].arguments
        self.assertEqual(transferred_amount_zneo, balance_zneo)
        self.assertEqual(transferred_amount_zgas, balance_zgas)

        mint_events = engine.get_events('Mint', origin=amm_address)
        self.assertEqual(1, len(mint_events))
        self.assertEqual(3, len(mint_events[0].arguments))

        address, amount_zneo, amount_zgas = mint_events[0].arguments
        if isinstance(address, str):
            address = String(address).to_bytes()
        self.assertEqual(self.OTHER_ACCOUNT_1, address)
        self.assertEqual(transferred_amount_zneo, amount_zneo)
        self.assertEqual(transferred_amount_zgas, amount_zgas)

        # data that will be compared with the previously saved data
        total_supply_after = self.run_smart_contract(engine, path, 'totalSupply')
        balance_user_amm_after = self.run_smart_contract(engine, path, 'balanceOf', self.OTHER_ACCOUNT_1)
        reserves_after = self.run_smart_contract(engine, path, 'get_reserves')
        balance_user_zneo_after = self.run_smart_contract(engine, path_zneo, 'balanceOf', self.OTHER_ACCOUNT_1)
        balance_user_zgas_after = self.run_smart_contract(engine, path_zgas, 'balanceOf', self.OTHER_ACCOUNT_1)
        balance_amm_zneo_after = self.run_smart_contract(engine, path_zneo, 'balanceOf', amm_address)
        balance_amm_zgas_after = self.run_smart_contract(engine, path_zgas, 'balanceOf', amm_address)

        self.assertEqual(total_supply_before + liquidity, total_supply_after)
        self.assertEqual(balance_user_amm_before + liquidity, balance_user_amm_after)
        self.assertEqual(reserves_before[0] + transferred_amount_zneo, reserves_after[0])
        self.assertEqual(reserves_before[1] + transferred_amount_zgas, reserves_after[1])
        self.assertEqual(balance_user_zneo_before - transferred_amount_zneo, balance_user_zneo_after)
        self.assertEqual(balance_user_zgas_before - transferred_amount_zgas, balance_user_zgas_after)
        self.assertEqual(reserves_before[0], balance_amm_zneo_before)
        self.assertEqual(reserves_before[1], balance_amm_zgas_before)
        self.assertEqual(reserves_after[0], balance_amm_zneo_after)
        self.assertEqual(reserves_after[1], balance_amm_zgas_after)

        transferred_amount_zneo = 2 * 10 ** 8
        transferred_amount_zgas = 23 * 10 ** 8

        # approving the AMM contract, so that it will be able to transfer zNEO from OTHER_ACCOUNT_1
        result = self.run_smart_contract(engine, path_zneo, 'approve',
                                         amm_address, transferred_amount_zneo,
                                         calling_script_hash=self.OTHER_ACCOUNT_1,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(True, result)

        # approving the AMM contract, so that it will be able to transfer zGAS from OTHER_ACCOUNT_1
        result = self.run_smart_contract(engine, path_zgas, 'approve',
                                         amm_address, transferred_amount_zgas,
                                         calling_script_hash=self.OTHER_ACCOUNT_1,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(True, result)

        # saving data to demonstrate that the value will change later
        total_supply_before = self.run_smart_contract(engine, path, 'totalSupply')
        balance_user_before = self.run_smart_contract(engine, path, 'balanceOf', self.OTHER_ACCOUNT_1)
        reserves_before = self.run_smart_contract(engine, path, 'get_reserves')
        balance_user_zneo_before = self.run_smart_contract(engine, path_zneo, 'balanceOf', self.OTHER_ACCOUNT_1)
        balance_user_zgas_before = self.run_smart_contract(engine, path_zgas, 'balanceOf', self.OTHER_ACCOUNT_1)
        balance_amm_zneo_before = self.run_smart_contract(engine, path_zneo, 'balanceOf', amm_address)
        balance_amm_zgas_before = self.run_smart_contract(engine, path_zgas, 'balanceOf', amm_address)
        # adding liquidity to the pool will give you AMM tokens in return
        result = self.run_smart_contract(engine, path, 'add_liquidity',
                                         transferred_amount_zneo, transferred_amount_zgas, 0, 0, self.OTHER_ACCOUNT_1,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        # zGAS will be quoted to keep the same ratio between zNEO and zGAS, the current ratio is 1 NEO to 11 GAS,
        # therefore, if 2 NEO are being added to the AMM, 22 GAS will be added instead of 23
        transferred_amount_zgas_quoted = transferred_amount_zneo * reserves_before[1] // reserves_before[0]
        # since there are tokens in the pool already, liquidity will be calculated as follows
        liquidity = min(transferred_amount_zneo * total_supply_before // reserves_before[0],
                        transferred_amount_zgas_quoted * total_supply_before // reserves_before[1])
        self.assertEqual(3, len(result))
        self.assertEqual(transferred_amount_zneo, result[0])
        self.assertEqual(transferred_amount_zgas_quoted, result[1])
        self.assertEqual(liquidity, result[2])

        transfer_events = engine.get_events('Transfer', origin=amm_address)
        self.assertEqual(2, len(transfer_events))
        self.assertEqual(3, len(transfer_events[1].arguments))

        sender, receiver, amount = transfer_events[1].arguments
        if isinstance(sender, str):
            sender = String(sender).to_bytes()
        if isinstance(receiver, str):
            receiver = String(receiver).to_bytes()
        self.assertEqual(None, sender)
        self.assertEqual(self.OTHER_ACCOUNT_1, receiver)
        self.assertEqual(liquidity, amount)

        sync_events = engine.get_events('Sync', origin=amm_address)
        self.assertEqual(2, len(sync_events))
        self.assertEqual(2, len(sync_events[1].arguments))

        balance_zneo, balance_zgas = sync_events[1].arguments
        self.assertEqual(reserves_before[0] + transferred_amount_zneo, balance_zneo)
        self.assertEqual(reserves_before[1] + transferred_amount_zgas_quoted, balance_zgas)

        mint_events = engine.get_events('Mint', origin=amm_address)
        self.assertEqual(2, len(mint_events))
        self.assertEqual(3, len(mint_events[1].arguments))

        address, amount_zneo, amount_zgas = mint_events[1].arguments
        if isinstance(address, str):
            address = String(address).to_bytes()
        self.assertEqual(self.OTHER_ACCOUNT_1, address)
        self.assertEqual(transferred_amount_zneo, amount_zneo)
        self.assertEqual(transferred_amount_zgas_quoted, amount_zgas)

        # data that will be compared with the previously saved data
        total_supply_after = self.run_smart_contract(engine, path, 'totalSupply')
        balance_user_after = self.run_smart_contract(engine, path, 'balanceOf', self.OTHER_ACCOUNT_1)
        reserves_after = self.run_smart_contract(engine, path, 'get_reserves')
        balance_user_zneo_after = self.run_smart_contract(engine, path_zneo, 'balanceOf', self.OTHER_ACCOUNT_1)
        balance_user_zgas_after = self.run_smart_contract(engine, path_zgas, 'balanceOf', self.OTHER_ACCOUNT_1)
        balance_amm_zneo_after = self.run_smart_contract(engine, path_zneo, 'balanceOf', amm_address)
        balance_amm_zgas_after = self.run_smart_contract(engine, path_zgas, 'balanceOf', amm_address)

        self.assertEqual(total_supply_before + liquidity, total_supply_after)
        self.assertEqual(balance_user_before + liquidity, balance_user_after)
        self.assertEqual(reserves_before[0] + transferred_amount_zneo, reserves_after[0])
        self.assertEqual(reserves_before[1] + transferred_amount_zgas_quoted, reserves_after[1])
        self.assertEqual(balance_user_zneo_before - transferred_amount_zneo, balance_user_zneo_after)
        self.assertEqual(balance_user_zgas_before - transferred_amount_zgas_quoted, balance_user_zgas_after)
        self.assertEqual(reserves_before[0], balance_amm_zneo_before)
        self.assertEqual(reserves_before[1], balance_amm_zgas_before)
        self.assertEqual(reserves_after[0], balance_amm_zneo_after)
        self.assertEqual(reserves_after[1], balance_amm_zgas_after)

    def test_amm_remove_liquidity(self):
        transferred_amount_zneo = 10 * 10 ** 8
        transferred_amount_zgas = 110 * 10 ** 8

        path = self.get_contract_path('amm.py')
        path_zneo = self.get_contract_path('wrapped_neo.py')
        path_zgas = self.get_contract_path('wrapped_gas.py')
        engine = TestEngine()

        self.run_smart_contract(engine, path, 'symbol')
        amm_address = engine.executed_script_hash.to_array()

        self.run_smart_contract(engine, path_zneo, 'symbol')
        zneo_address = engine.executed_script_hash.to_array()

        self.run_smart_contract(engine, path_zgas, 'symbol')
        zgas_address = engine.executed_script_hash.to_array()

        engine.reset_engine()
        engine.add_contract(path.replace('.py', '.nef'))
        engine.add_contract(path_zneo.replace('.py', '.nef'))
        engine.add_contract(path_zgas.replace('.py', '.nef'))

        result = self.run_smart_contract(engine, path, 'set_address', zneo_address, zgas_address,
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(True, result)

        # can't remove liquidity, because the user doesn't have any
        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'remove_liquidity', 10000, 0, 0, self.OTHER_ACCOUNT_1,
                                    signer_accounts=[self.OWNER_SCRIPT_HASH])

        engine.add_neo(self.OTHER_ACCOUNT_1, 10_000_000 * 10 ** 8)
        engine.add_gas(self.OTHER_ACCOUNT_1, 10_000_000 * 10 ** 8)

        # minting zNEO to OTHER_ACCOUNT_1
        result = self.run_smart_contract(engine, constants.NEO_SCRIPT, 'transfer',
                                         self.OTHER_ACCOUNT_1, zneo_address, 10_000_000 * 10 ** 8, None,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(True, result)

        # minting zGAS to OTHER_ACCOUNT_1
        result = self.run_smart_contract(engine, constants.GAS_SCRIPT, 'transfer',
                                         self.OTHER_ACCOUNT_1, zgas_address, 10_000_000 * 10 ** 8, None,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(True, result)

        # approving the AMM contract, so that it will be able to transfer zNEO from OTHER_ACCOUNT_1
        result = self.run_smart_contract(engine, path_zneo, 'approve',
                                         amm_address, transferred_amount_zneo,
                                         calling_script_hash=self.OTHER_ACCOUNT_1,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(True, result)

        # approving the AMM contract, so that it will be able to transfer zGAS from OTHER_ACCOUNT_1
        result = self.run_smart_contract(engine, path_zgas, 'approve',
                                         amm_address, transferred_amount_zgas,
                                         calling_script_hash=self.OTHER_ACCOUNT_1,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(True, result)

        # adding liquidity to the pool will give you AMM tokens in return
        result = self.run_smart_contract(engine, path, 'add_liquidity',
                                         transferred_amount_zneo, transferred_amount_zgas, 0, 0, self.OTHER_ACCOUNT_1,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        import math
        liquidity = int(math.sqrt(transferred_amount_zneo * transferred_amount_zgas))
        self.assertEqual(3, len(result))
        self.assertEqual(transferred_amount_zneo, result[0])
        self.assertEqual(transferred_amount_zgas, result[1])
        self.assertEqual(liquidity, result[2])

        # saving data to demonstrate that the value will change later
        total_supply_before = self.run_smart_contract(engine, path, 'totalSupply')
        balance_user_before = self.run_smart_contract(engine, path, 'balanceOf', self.OTHER_ACCOUNT_1)
        reserves_before = self.run_smart_contract(engine, path, 'get_reserves')
        balance_user_zneo_before = self.run_smart_contract(engine, path_zneo, 'balanceOf', self.OTHER_ACCOUNT_1)
        balance_user_zgas_before = self.run_smart_contract(engine, path_zgas, 'balanceOf', self.OTHER_ACCOUNT_1)
        balance_amm_zneo_before = self.run_smart_contract(engine, path_zneo, 'balanceOf', amm_address)
        balance_amm_zgas_before = self.run_smart_contract(engine, path_zgas, 'balanceOf', amm_address)
        # removing liquidity from the pool will return the equivalent zNEO and zGAS that were used to fund the pool
        result = self.run_smart_contract(engine, path, 'remove_liquidity', liquidity, 0, 0, self.OTHER_ACCOUNT_1,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(2, len(result))
        self.assertEqual(transferred_amount_zneo, result[0])
        self.assertEqual(transferred_amount_zgas, result[1])

        transfer_events = engine.get_events('Transfer', origin=amm_address)
        # add_liquidity sent a Transfer event and remove_liquidity sent another
        self.assertEqual(2, len(transfer_events))
        self.assertEqual(3, len(transfer_events[1].arguments))

        sender, receiver, amount = transfer_events[1].arguments
        if isinstance(sender, str):
            sender = String(sender).to_bytes()
        if isinstance(receiver, str):
            receiver = String(receiver).to_bytes()
        self.assertEqual(self.OTHER_ACCOUNT_1, sender)
        self.assertEqual(None, receiver)
        self.assertEqual(liquidity, amount)

        sync_events = engine.get_events('Sync', origin=amm_address)
        self.assertEqual(2, len(sync_events))
        self.assertEqual(2, len(sync_events[1].arguments))

        balance_zneo, balance_zgas = sync_events[1].arguments
        self.assertEqual(reserves_before[0] - transferred_amount_zneo, balance_zneo)
        self.assertEqual(reserves_before[1] - transferred_amount_zgas, balance_zgas)

        burn_events = engine.get_events('Burn', origin=amm_address)
        self.assertEqual(1, len(burn_events))
        self.assertEqual(3, len(burn_events[0].arguments))

        address, amount_zneo, amount_zgas = burn_events[0].arguments
        if isinstance(address, str):
            address = String(address).to_bytes()
        self.assertEqual(self.OTHER_ACCOUNT_1, address)
        self.assertEqual(transferred_amount_zneo, amount_zneo)
        self.assertEqual(transferred_amount_zgas, amount_zgas)

        # data that will be compared with the previously saved data
        total_supply_after = self.run_smart_contract(engine, path, 'totalSupply')
        balance_user_after = self.run_smart_contract(engine, path, 'balanceOf', self.OTHER_ACCOUNT_1)
        reserves_after = self.run_smart_contract(engine, path, 'get_reserves')
        balance_user_zneo_after = self.run_smart_contract(engine, path_zneo, 'balanceOf', self.OTHER_ACCOUNT_1)
        balance_user_zgas_after = self.run_smart_contract(engine, path_zgas, 'balanceOf', self.OTHER_ACCOUNT_1)
        balance_amm_zneo_after = self.run_smart_contract(engine, path_zneo, 'balanceOf', amm_address)
        balance_amm_zgas_after = self.run_smart_contract(engine, path_zgas, 'balanceOf', amm_address)

        self.assertEqual(total_supply_before - liquidity, total_supply_after)
        self.assertEqual(balance_user_before - liquidity, balance_user_after)
        self.assertEqual(reserves_before[0] - transferred_amount_zneo, reserves_after[0])
        self.assertEqual(reserves_before[1] - transferred_amount_zgas, reserves_after[1])
        self.assertEqual(balance_user_zneo_before + transferred_amount_zneo, balance_user_zneo_after)
        self.assertEqual(balance_user_zgas_before + transferred_amount_zgas, balance_user_zgas_after)
        self.assertEqual(reserves_before[0], balance_amm_zneo_before)
        self.assertEqual(reserves_before[1], balance_amm_zgas_before)
        self.assertEqual(reserves_after[0], balance_amm_zneo_after)
        self.assertEqual(reserves_after[1], balance_amm_zgas_after)

    def test_amm_swap_zneo_to_zgas(self):
        transferred_amount_zneo = 10 * 10 ** 8
        transferred_amount_zgas = 110 * 10 ** 8

        path = self.get_contract_path('amm.py')
        path_zneo = self.get_contract_path('wrapped_neo.py')
        path_zgas = self.get_contract_path('wrapped_gas.py')
        engine = TestEngine()

        self.run_smart_contract(engine, path, 'symbol')
        amm_address = engine.executed_script_hash.to_array()

        self.run_smart_contract(engine, path_zneo, 'symbol')
        zneo_address = engine.executed_script_hash.to_array()

        self.run_smart_contract(engine, path_zgas, 'symbol')
        zgas_address = engine.executed_script_hash.to_array()

        engine.reset_engine()
        engine.add_contract(path.replace('.py', '.nef'))
        engine.add_contract(path_zneo.replace('.py', '.nef'))
        engine.add_contract(path_zgas.replace('.py', '.nef'))

        result = self.run_smart_contract(engine, path, 'set_address', zneo_address, zgas_address,
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(True, result)

        # transferring zNEO to OTHER_ACCOUNT_1
        result = self.run_smart_contract(engine, path_zneo, 'transfer', self.OWNER_SCRIPT_HASH, self.OTHER_ACCOUNT_1,
                                         10_000_000 * 10 ** 8, None,
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(True, result)

        # transferring zGAS to OTHER_ACCOUNT_1
        result = self.run_smart_contract(engine, path_zgas, 'transfer', self.OWNER_SCRIPT_HASH, self.OTHER_ACCOUNT_1,
                                         10_000_000 * 10 ** 8, None,
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(True, result)

        # approving the AMM contract, so that it will be able to transfer zNEO from OTHER_ACCOUNT_1
        result = self.run_smart_contract(engine, path_zneo, 'approve',
                                         amm_address, transferred_amount_zneo,
                                         calling_script_hash=self.OTHER_ACCOUNT_1,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(True, result)

        # approving the AMM contract, so that it will be able to transfer zGAS from OTHER_ACCOUNT_1
        result = self.run_smart_contract(engine, path_zgas, 'approve',
                                         amm_address, transferred_amount_zgas,
                                         calling_script_hash=self.OTHER_ACCOUNT_1,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(True, result)

        # adding liquidity to the pool will give you AMM tokens in return
        result = self.run_smart_contract(engine, path, 'add_liquidity',
                                         transferred_amount_zneo, transferred_amount_zgas, 0, 0, self.OTHER_ACCOUNT_1,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        import math
        liquidity = int(math.sqrt(transferred_amount_zneo * transferred_amount_zgas))
        self.assertEqual(3, len(result))
        self.assertEqual(transferred_amount_zneo, result[0])
        self.assertEqual(transferred_amount_zgas, result[1])
        self.assertEqual(liquidity, result[2])

        swapped_zneo = 1 * 10 ** 8
        # won't work, because user did not enough zNEO tokens
        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'swap_tokens',
                                    swapped_zneo, 0, zneo_address, self.OTHER_ACCOUNT_1,
                                    signer_accounts=[self.OTHER_ACCOUNT_1])

        # approving the AMM contract, so that it will be able to transfer zNEO from OTHER_ACCOUNT_1
        result = self.run_smart_contract(engine, path_zneo, 'approve',
                                         amm_address, swapped_zneo,
                                         calling_script_hash=self.OTHER_ACCOUNT_1,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(True, result)

        # saving data to demonstrate that the value will change later
        total_supply_before = self.run_smart_contract(engine, path, 'totalSupply')
        reserves_before = self.run_smart_contract(engine, path, 'get_reserves')
        balance_user_zneo_before = self.run_smart_contract(engine, path_zneo, 'balanceOf', self.OTHER_ACCOUNT_1)
        balance_user_zgas_before = self.run_smart_contract(engine, path_zgas, 'balanceOf', self.OTHER_ACCOUNT_1)
        balance_amm_zneo_before = self.run_smart_contract(engine, path_zneo, 'balanceOf', amm_address)
        balance_amm_zgas_before = self.run_smart_contract(engine, path_zgas, 'balanceOf', amm_address)
        # swapping zneo for zgas
        result = self.run_smart_contract(engine, path, 'swap_tokens',
                                         swapped_zneo, 0, zneo_address, self.OTHER_ACCOUNT_1,
                                         signer_accounts=[self.OTHER_ACCOUNT_1], rollback_on_fault=False)
        # there is a 0.3% fee when doing a swap
        swapped_zneo_with_fee = swapped_zneo * (1000 - 3)
        swapped_zgas = swapped_zneo_with_fee * reserves_before[1] // (reserves_before[0] * 1000 + swapped_zneo_with_fee)
        self.assertEqual(swapped_zgas, result)

        # add_liquidity sent a Sync before
        transfer_events = engine.get_events('Sync')
        self.assertEqual(2, len(transfer_events))
        self.assertEqual(2, len(transfer_events[1].arguments))

        balance_zneo, balance_zgas = transfer_events[1].arguments
        self.assertEqual(reserves_before[0] + swapped_zneo, balance_zneo)
        self.assertEqual(reserves_before[1] - swapped_zgas, balance_zgas)

        transfer_events = engine.get_events('Swap')
        self.assertEqual(1, len(transfer_events))
        self.assertEqual(5, len(transfer_events[0].arguments))

        address, amount_zneo_in, amount_zgas_in, amount_zneo_out, amount_zgas_out = transfer_events[0].arguments
        if isinstance(address, str):
            address = String(address).to_bytes()
        self.assertEqual(self.OTHER_ACCOUNT_1, address)
        self.assertEqual(swapped_zneo, amount_zneo_in)
        self.assertEqual(0, amount_zgas_in)
        self.assertEqual(0, amount_zneo_out)
        self.assertEqual(swapped_zgas, amount_zgas_out)

        # data that will be compared with the previously saved data
        total_supply_after = self.run_smart_contract(engine, path, 'totalSupply')
        reserves_after = self.run_smart_contract(engine, path, 'get_reserves')
        balance_user_zneo_after = self.run_smart_contract(engine, path_zneo, 'balanceOf', self.OTHER_ACCOUNT_1)
        balance_user_zgas_after = self.run_smart_contract(engine, path_zgas, 'balanceOf', self.OTHER_ACCOUNT_1)
        balance_amm_zneo_after = self.run_smart_contract(engine, path_zneo, 'balanceOf', amm_address)
        balance_amm_zgas_after = self.run_smart_contract(engine, path_zgas, 'balanceOf', amm_address)

        self.assertEqual(total_supply_before, total_supply_after)
        self.assertEqual(reserves_before[0] + swapped_zneo, reserves_after[0])
        self.assertEqual(reserves_before[1] - swapped_zgas, reserves_after[1])
        self.assertEqual(balance_user_zneo_before - swapped_zneo, balance_user_zneo_after)
        self.assertEqual(balance_user_zgas_before + swapped_zgas, balance_user_zgas_after)
        self.assertEqual(reserves_before[0], balance_amm_zneo_before)
        self.assertEqual(reserves_before[1], balance_amm_zgas_before)
        self.assertEqual(reserves_after[0], balance_amm_zneo_after)
        self.assertEqual(reserves_after[1], balance_amm_zgas_after)
        self.assertEqual(reserves_before[0] + swapped_zneo, reserves_after[0])
        self.assertEqual(reserves_before[1] - swapped_zgas, reserves_after[1])

    def test_amm_swap_zgas_to_zneo(self):
        transferred_amount_zneo = 10 * 10 ** 8
        transferred_amount_zgas = 110 * 10 ** 8

        path = self.get_contract_path('amm.py')
        path_zneo = self.get_contract_path('wrapped_neo.py')
        path_zgas = self.get_contract_path('wrapped_gas.py')
        engine = TestEngine()

        self.run_smart_contract(engine, path, 'symbol')
        amm_address = engine.executed_script_hash.to_array()

        self.run_smart_contract(engine, path_zneo, 'symbol')
        zneo_address = engine.executed_script_hash.to_array()

        self.run_smart_contract(engine, path_zgas, 'symbol')
        zgas_address = engine.executed_script_hash.to_array()

        engine.reset_engine()
        engine.add_contract(path.replace('.py', '.nef'))
        engine.add_contract(path_zneo.replace('.py', '.nef'))
        engine.add_contract(path_zgas.replace('.py', '.nef'))

        result = self.run_smart_contract(engine, path, 'set_address', zneo_address, zgas_address,
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(True, result)

        # transferring zNEO to OTHER_ACCOUNT_1
        result = self.run_smart_contract(engine, path_zneo, 'transfer', self.OWNER_SCRIPT_HASH, self.OTHER_ACCOUNT_1,
                                         10_000_000 * 10 ** 8, None,
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(True, result)

        # transferring zGAS to OTHER_ACCOUNT_1
        result = self.run_smart_contract(engine, path_zgas, 'transfer', self.OWNER_SCRIPT_HASH, self.OTHER_ACCOUNT_1,
                                         10_000_000 * 10 ** 8, None,
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(True, result)

        # approving the AMM contract, so that it will be able to transfer zNEO from OTHER_ACCOUNT_1
        result = self.run_smart_contract(engine, path_zneo, 'approve',
                                         amm_address, transferred_amount_zneo,
                                         calling_script_hash=self.OTHER_ACCOUNT_1,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(True, result)

        # approving the AMM contract, so that it will be able to transfer zGAS from OTHER_ACCOUNT_1
        result = self.run_smart_contract(engine, path_zgas, 'approve',
                                         amm_address, transferred_amount_zgas,
                                         calling_script_hash=self.OTHER_ACCOUNT_1,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(True, result)

        # adding liquidity to the pool will give you AMM tokens in return
        result = self.run_smart_contract(engine, path, 'add_liquidity',
                                         transferred_amount_zneo, transferred_amount_zgas, 0, 0, self.OTHER_ACCOUNT_1,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        import math
        liquidity = int(math.sqrt(transferred_amount_zneo * transferred_amount_zgas))
        self.assertEqual(3, len(result))
        self.assertEqual(transferred_amount_zneo, result[0])
        self.assertEqual(transferred_amount_zgas, result[1])
        self.assertEqual(liquidity, result[2])

        swapped_zgas = 11 * 10 ** 8
        # won't work, because user did not enough zNEO tokens
        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'swap_tokens',
                                    swapped_zgas, 0, zgas_address, self.OTHER_ACCOUNT_1,
                                    signer_accounts=[self.OTHER_ACCOUNT_1])

        # approving the AMM contract, so that it will be able to transfer zNEO from OTHER_ACCOUNT_1
        result = self.run_smart_contract(engine, path_zgas, 'approve',
                                         amm_address, swapped_zgas,
                                         calling_script_hash=self.OTHER_ACCOUNT_1,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(True, result)

        # saving data to demonstrate that the value will change later
        total_supply_before = self.run_smart_contract(engine, path, 'totalSupply')
        reserves_before = self.run_smart_contract(engine, path, 'get_reserves')
        balance_user_zneo_before = self.run_smart_contract(engine, path_zneo, 'balanceOf', self.OTHER_ACCOUNT_1)
        balance_user_zgas_before = self.run_smart_contract(engine, path_zgas, 'balanceOf', self.OTHER_ACCOUNT_1)
        balance_amm_zneo_before = self.run_smart_contract(engine, path_zneo, 'balanceOf', amm_address)
        balance_amm_zgas_before = self.run_smart_contract(engine, path_zgas, 'balanceOf', amm_address)
        # swapping zgas for zneo
        result = self.run_smart_contract(engine, path, 'swap_tokens',
                                         swapped_zgas, 0, zgas_address, self.OTHER_ACCOUNT_1,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        # there is a 0.3% fee when doing a swap
        swapped_zgas_with_fee = swapped_zgas * (1000 - 3)
        swapped_zneo = swapped_zgas_with_fee * reserves_before[0] // (reserves_before[1] * 1000 + swapped_zgas_with_fee)
        self.assertEqual(swapped_zneo, result)

        # add_liquidity sent a Sync before
        transfer_events = engine.get_events('Sync')
        self.assertEqual(2, len(transfer_events))
        self.assertEqual(2, len(transfer_events[1].arguments))

        balance_zneo, balance_zgas = transfer_events[1].arguments
        self.assertEqual(reserves_before[0] - swapped_zneo, balance_zneo)
        self.assertEqual(reserves_before[1] + swapped_zgas, balance_zgas)

        transfer_events = engine.get_events('Swap')
        self.assertEqual(1, len(transfer_events))
        self.assertEqual(5, len(transfer_events[0].arguments))

        address, amount_zneo_in, amount_zgas_in, amount_zneo_out, amount_zgas_out = transfer_events[0].arguments
        if isinstance(address, str):
            address = String(address).to_bytes()
        self.assertEqual(self.OTHER_ACCOUNT_1, address)
        self.assertEqual(0, amount_zneo_in)
        self.assertEqual(swapped_zgas, amount_zgas_in)
        self.assertEqual(swapped_zneo, amount_zneo_out)
        self.assertEqual(0, amount_zgas_out)

        # data that will be compared with the previously saved data
        total_supply_after = self.run_smart_contract(engine, path, 'totalSupply')
        reserves_after = self.run_smart_contract(engine, path, 'get_reserves')
        balance_user_zneo_after = self.run_smart_contract(engine, path_zneo, 'balanceOf', self.OTHER_ACCOUNT_1)
        balance_user_zgas_after = self.run_smart_contract(engine, path_zgas, 'balanceOf', self.OTHER_ACCOUNT_1)
        balance_amm_zneo_after = self.run_smart_contract(engine, path_zneo, 'balanceOf', amm_address)
        balance_amm_zgas_after = self.run_smart_contract(engine, path_zgas, 'balanceOf', amm_address)

        self.assertEqual(total_supply_before, total_supply_after)
        self.assertEqual(reserves_before[0] - swapped_zneo, reserves_after[0])
        self.assertEqual(reserves_before[1] + swapped_zgas, reserves_after[1])
        self.assertEqual(balance_user_zneo_before + swapped_zneo, balance_user_zneo_after)
        self.assertEqual(balance_user_zgas_before - swapped_zgas, balance_user_zgas_after)
        self.assertEqual(reserves_before[0], balance_amm_zneo_before)
        self.assertEqual(reserves_before[1], balance_amm_zgas_before)
        self.assertEqual(reserves_after[0], balance_amm_zneo_after)
        self.assertEqual(reserves_after[1], balance_amm_zgas_after)
        self.assertEqual(reserves_before[0] - swapped_zneo, reserves_after[0])
        self.assertEqual(reserves_before[1] + swapped_zgas, reserves_after[1])



from boa3.boa3 import Boa3
from boa3.exception import CompilerError
from boa3.model.operation.binaryop import BinaryOp
from boa3.model.type.type import Type
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3.neo3.contracts import FindOptions
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.TestExecutionException import TestExecutionException
from boa3_test.tests.test_classes.testengine import TestEngine


class TestArithmetic(BoaTest):
    default_folder: str = 'test_sc/arithmetic_test'

    # region Addition

    def test_boa2_add_test(self):
        path = self.get_contract_path('AddBoa2Test.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main', 2)
        self.assertEqual(4, result)

        result = self.run_smart_contract(engine, path, 'main', 23234)
        self.assertEqual(23236, result)

        result = self.run_smart_contract(engine, path, 'main', 0)
        self.assertEqual(2, result)

        result = self.run_smart_contract(engine, path, 'main', -112)
        self.assertEqual(-110, result)

    def test_boa2_add_test1(self):
        path = self.get_contract_path('AddBoa2Test1.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main', 1, 2, 3, 4)
        self.assertEqual(9, result)

        result = self.run_smart_contract(engine, path, 'main', 0, 0, 0, 2)
        self.assertEqual(2, result)

        result = self.run_smart_contract(engine, path, 'main', -2, 3, -6, 2)
        self.assertEqual(-2, result)

    def test_boa2_add_test2(self):
        path = self.get_contract_path('AddBoa2Test2.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(3, result)

    def test_boa2_add_test3(self):
        path = self.get_contract_path('AddBoa2Test3.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(-9, result)

    def test_boa2_add_test4(self):
        path = self.get_contract_path('AddBoa2Test4.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main', 1, 2, 3, 4)
        self.assertEqual(-9, result)

    def test_boa2_add_test_void(self):
        path = self.get_contract_path('AddBoa2TestVoid.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main', 3)
        self.assertIsVoid(result)

    def test_addition_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.ADD
            + Opcode.RET
        )

        path = self.get_contract_path('Addition.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'add', 1, 2)
        self.assertEqual(3, result)
        result = self.run_smart_contract(engine, path, 'add', -42, -24)
        self.assertEqual(-66, result)
        result = self.run_smart_contract(engine, path, 'add', -42, 24)
        self.assertEqual(-18, result)

    def test_addition_augmented_assignment(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.ADD
            + Opcode.STARG0
            + Opcode.RET
        )

        path = self.get_contract_path('AdditionAugmentedAssignment.py')
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_addition_builtin_type(self):
        path = self.get_contract_path('AdditionBuiltinType.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', FindOptions.VALUES_ONLY, FindOptions.DESERIALIZE_VALUES)
        self.assertEqual(FindOptions.VALUES_ONLY + FindOptions.DESERIALIZE_VALUES, result)

    def test_addition_literal_operation(self):
        expected_output = (
            Opcode.PUSH3
            + Opcode.RET
        )

        path = self.get_contract_path('AdditionLiteral.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'add_one_two')
        self.assertEqual(3, result)

    def test_addition_literal_and_variable(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x01'
            + Opcode.PUSH1
            + Opcode.LDARG0
            + Opcode.ADD
            + Opcode.RET
        )

        path = self.get_contract_path('AdditionLiteralAndVariable.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'add_one', 1)
        self.assertEqual(2, result)
        result = self.run_smart_contract(engine, path, 'add_one', -10)
        self.assertEqual(-9, result)
        result = self.run_smart_contract(engine, path, 'add_one', -1)
        self.assertEqual(0, result)

    def test_sequence_addition(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.ADD
            + Opcode.PUSH4
            + Opcode.ADD
            + Opcode.RET
        )

        path = self.get_contract_path('AdditionThreeElements.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'add_four', 1, 2)
        self.assertEqual(7, result)
        result = self.run_smart_contract(engine, path, 'add_four', -42, -24)
        self.assertEqual(-62, result)
        result = self.run_smart_contract(engine, path, 'add_four', -42, 24)
        self.assertEqual(-14, result)

    def test_sequence_addition_different_orders(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00\x01'
            + Opcode.PUSH6
            + Opcode.LDARG0
            + Opcode.ADD
            + Opcode.RET
        )
        path_1 = self.get_contract_path('AdditionThreeValuesUnordered1.py')
        output_1 = Boa3.compile(path_1)
        self.assertEqual(expected_output, output_1)

        path_2 = self.get_contract_path('AdditionThreeValuesUnordered2.py')
        output_2 = Boa3.compile(path_2)
        self.assertEqual(expected_output, output_2)

        path_3 = self.get_contract_path('AdditionThreeValuesUnordered3.py')
        output_3 = Boa3.compile(path_3)
        self.assertEqual(expected_output, output_3)

        engine = TestEngine()
        result_1 = self.run_smart_contract(engine, path_1, 'add_six', 5)
        result_2 = self.run_smart_contract(engine, path_2, 'add_six', 5)
        result_3 = self.run_smart_contract(engine, path_2, 'add_six', 5)
        self.assertEqual(11, result_1)
        self.assertEqual(11, result_2)
        self.assertEqual(11, result_3)

        result_1 = self.run_smart_contract(engine, path_1, 'add_six', -42)
        result_2 = self.run_smart_contract(engine, path_2, 'add_six', -42)
        result_3 = self.run_smart_contract(engine, path_2, 'add_six', -42)
        self.assertEqual(-36, result_1)
        self.assertEqual(-36, result_2)
        self.assertEqual(-36, result_3)

    def test_addition_variable_and_literal(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0
            + Opcode.PUSH1
            + Opcode.ADD
            + Opcode.RET
        )

        path = self.get_contract_path('AdditionVariableAndLiteral.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'add_one', 1)
        self.assertEqual(2, result)
        result = self.run_smart_contract(engine, path, 'add_one', -10)
        self.assertEqual(-9, result)
        result = self.run_smart_contract(engine, path, 'add_one', -1)
        self.assertEqual(0, result)

    # endregion

    # region Concatenation

    def test_concat_bytes_variables_and_constants(self):
        path = self.get_contract_path('ConcatBytesVariablesAndConstants.py')

        engine = TestEngine()
        engine.increase_block()  # increase to get consistent block_time between execution and engine
        result = self.run_smart_contract(engine, path, 'concat1',
                                         expected_result_type=bytes)
        current_time = Integer(engine.current_block.timestamp).to_byte_array()
        self.assertEqual(b'value1  value2  value3  ' + current_time + b'some_bytes_after', result)

        result = self.run_smart_contract(engine, path, 'concat2',
                                         expected_result_type=bytes)
        current_time = Integer(engine.current_block.timestamp).to_byte_array()
        self.assertEqual(b'value1value2value3' + current_time + b'some_bytes_after', result)

        result = self.run_smart_contract(engine, path, 'concat3',
                                         expected_result_type=bytes)
        current_time = Integer(engine.current_block.timestamp).to_byte_array()
        self.assertEqual(b'value1__value2__value3__' + current_time + b'some_bytes_after', result)

    def test_concatenation_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.CAT
            + Opcode.CONVERT
            + Type.str.stack_item
            + Opcode.RET
        )

        path = self.get_contract_path('Concatenation.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'concat', 'a', 'b')
        self.assertEqual('ab', result)
        result = self.run_smart_contract(engine, path, 'concat', 'unit', 'test')
        self.assertEqual('unittest', result)

    def test_concatenation_augmented_assignment(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.CAT
            + Opcode.CONVERT
            + Type.str.stack_item
            + Opcode.STARG0
            + Opcode.RET
        )

        path = self.get_contract_path('ConcatenationAugmentedAssignment.py')
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_concat_string_variables_and_constants(self):
        path = self.get_contract_path('ConcatStringVariablesAndConstants.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'concat')
        self.assertEqual('[1,2]', result)

    # endregion

    # region Division

    def test_division_operation(self):
        path = self.get_contract_path('Division.py')
        self.assertCompilerLogs(CompilerError.NotSupportedOperation, path)

    def test_division_augmented_assignment(self):
        path = self.get_contract_path('DivisionAugmentedAssignment.py')
        self.assertCompilerLogs(CompilerError.NotSupportedOperation, path)

    def test_division_builtin_type(self):
        path = self.get_contract_path('DivisionBuiltinType.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', FindOptions.DESERIALIZE_VALUES, FindOptions.VALUES_ONLY)
        self.assertEqual(FindOptions.DESERIALIZE_VALUES // FindOptions.VALUES_ONLY, result)

    def test_integer_division_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.DIV
            + Opcode.RET
        )

        path = self.get_contract_path('IntegerDivision.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'floor_div', 10, 3)
        self.assertEqual(3, result)
        result = self.run_smart_contract(engine, path, 'floor_div', -42, -9)
        self.assertEqual(4, result)
        result = self.run_smart_contract(engine, path, 'floor_div', -100, 3)
        self.assertEqual(-33, result)

    def test_integer_division_augmented_assignment(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.DIV
            + Opcode.STARG0
            + Opcode.RET
        )

        path = self.get_contract_path('IntegerDivisionAugmentedAssignment.py')
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    # endregion

    # region ListAddition

    def test_list_addition(self):
        path = self.get_contract_path('ListAddition.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'add_any', [1, 'str', '123'], [2, True, False])
        self.assertEqual([1, 'str', '123'] + [2, True, False], result)
        result = self.run_smart_contract(engine, path, 'add_int', [1, 3], [2, 5])
        self.assertEqual([1, 3] + [2, 5], result)
        result = self.run_smart_contract(engine, path, 'add_bool', [True], [False, True])
        self.assertEqual([True] + [False, True], result)
        result = self.run_smart_contract(engine, path, 'add_str', ['unit', ' '], ['test', '.'])
        self.assertEqual(['unit', ' '] + ['test', '.'], result)

    # endregion

    # region Mismatched

    def test_mismatched_type_binary_operation(self):
        path = self.get_contract_path('MismatchedOperandBinary.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_mismatched_type_unary_operation(self):
        path = self.get_contract_path('MismatchedOperandUnary.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    # endregion

    # region Mixed

    def test_mixed_operations(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x05'
            + Opcode.LDARG0
            + Opcode.LDARG2
            + Opcode.LDARG4
            + Opcode.MUL        # multiplicative operations
            + Opcode.ADD        # additive operations
            + Opcode.LDARG3
            + Opcode.NEGATE     # parentheses
            + Opcode.LDARG1
            + Opcode.DIV        # multiplicative
            + Opcode.SUB        # additive
            + Opcode.RET
        )

        path = self.get_contract_path('MixedOperations.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'mixed', 10, 20, 30, 40, 50)
        self.assertEqual(10 + 30 * 50 + 40 // 20, result)

    def test_mixed_operations_with_parentheses(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x05'
            + Opcode.LDARG0
            + Opcode.LDARG2
            + Opcode.LDARG4
            + Opcode.LDARG3
            + Opcode.NEGATE     # inside parentheses
            + Opcode.SUB        # parentheses
            + Opcode.MUL        # multiplicative operations
            + Opcode.LDARG1
            + Opcode.DIV        # multiplicative
            + Opcode.ADD        # additive operations
            + Opcode.RET
        )

        path = self.get_contract_path('WithParentheses.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'mixed', 10, 20, 30, 40, 50)
        self.assertEqual(10 + 30 * (50 + 40) // 20, result)

    # endregion

    # region Modulo

    def test_modulo_operation(self):
        path = self.get_contract_path('Modulo.py')
        engine = TestEngine()

        op1 = 10
        op2 = 3
        expected_output = op1 % op2
        result = self.run_smart_contract(engine, path, 'mod', op1, op2)
        self.assertEqual(expected_output, result)

        op1 = -42
        op2 = -9
        expected_output = op1 % op2
        result = self.run_smart_contract(engine, path, 'mod', op1, op2)
        self.assertEqual(expected_output, result)

        op1 = -100
        op2 = 3
        expected_output = op1 % op2
        result = self.run_smart_contract(engine, path, 'mod', op1, op2)
        self.assertEqual(expected_output, result)

        op1 = 100
        op2 = -3
        expected_output = op1 % op2
        result = self.run_smart_contract(engine, path, 'mod', op1, op2)
        self.assertEqual(expected_output, result)

    def test_modulo_augmented_assignment(self):
        path = self.get_contract_path('ModuloAugmentedAssignment.py')
        engine = TestEngine()

        op1 = 10
        op2 = 3
        expected_output = op1 % op2
        result = self.run_smart_contract(engine, path, 'mod', op1, op2)
        self.assertEqual(expected_output, result)

        op1 = -42
        op2 = -9
        expected_output = op1 % op2
        result = self.run_smart_contract(engine, path, 'mod', op1, op2)
        self.assertEqual(expected_output, result)

        op1 = -100
        op2 = 3
        expected_output = op1 % op2
        result = self.run_smart_contract(engine, path, 'mod', op1, op2)
        self.assertEqual(expected_output, result)

        op1 = 100
        op2 = -3
        expected_output = op1 % op2
        result = self.run_smart_contract(engine, path, 'mod', op1, op2)
        self.assertEqual(expected_output, result)

    def test_modulo_builtin_type(self):
        path = self.get_contract_path('ModuloBuiltinType.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', FindOptions.DESERIALIZE_VALUES, FindOptions.VALUES_ONLY)
        self.assertEqual(FindOptions.DESERIALIZE_VALUES % FindOptions.VALUES_ONLY, result)

    # endregion

    # region Multiplication

    def test_multiplication_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.MUL
            + Opcode.RET
        )

        path = self.get_contract_path('Multiplication.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'mult', 10, 3)
        self.assertEqual(30, result)
        result = self.run_smart_contract(engine, path, 'mult', -42, -2)
        self.assertEqual(84, result)
        result = self.run_smart_contract(engine, path, 'mult', -4, 20)
        self.assertEqual(-80, result)
        result = self.run_smart_contract(engine, path, 'mult', 0, 20)
        self.assertEqual(0, result)

    def test_multiplication_augmented_assignment(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.MUL
            + Opcode.STARG0
            + Opcode.RET
        )

        path = self.get_contract_path('MultiplicationAugmentedAssignment.py')
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_multiplication_builtin_type(self):
        path = self.get_contract_path('MultiplicationBuiltinType.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', FindOptions.DESERIALIZE_VALUES, FindOptions.VALUES_ONLY)
        self.assertEqual(FindOptions.DESERIALIZE_VALUES * FindOptions.VALUES_ONLY, result)

    # endregion

    # region Power

    def test_power_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.POW
            + Opcode.RET
        )
        path = self.get_contract_path('Power.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'pow', 10, 3)
        self.assertEqual(1000, result)
        result = self.run_smart_contract(engine, path, 'pow', 1, 15)
        self.assertEqual(1, result)
        result = self.run_smart_contract(engine, path, 'pow', -2, 2)
        self.assertEqual(4, result)
        result = self.run_smart_contract(engine, path, 'pow', 0, 20)
        self.assertEqual(0, result)
        with self.assertRaisesRegex(TestExecutionException, '^Invalid shift value'):
            self.run_smart_contract(engine, path, 'pow', 1, -2)

    def test_power_augmented_assignment(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.POW
            + Opcode.STARG0
            + Opcode.RET
        )
        path = self.get_contract_path('PowerAugmentedAssignment.py')
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_power_builtin_type(self):
        path = self.get_contract_path('PowerBuiltinType.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', FindOptions.DESERIALIZE_VALUES, FindOptions.VALUES_ONLY)
        self.assertEqual(FindOptions.DESERIALIZE_VALUES ** FindOptions.VALUES_ONLY, result)

    # endregion

    # region Sign

    def test_negative_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0
            + Opcode.NEGATE
            + Opcode.RET
        )

        path = self.get_contract_path('Negative.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'minus', 10)
        self.assertEqual(-10, result)
        result = self.run_smart_contract(engine, path, 'minus', -1)
        self.assertEqual(1, result)
        result = self.run_smart_contract(engine, path, 'minus', 0)
        self.assertEqual(0, result)

    def test_negative_builtin_type(self):
        path = self.get_contract_path('NegativeBuiltinType.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'minus', FindOptions.DESERIALIZE_VALUES)
        self.assertEqual(-FindOptions.DESERIALIZE_VALUES, result)

    def test_positive_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0
            + Opcode.RET
        )

        path = self.get_contract_path('Positive.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'plus', 10)
        self.assertEqual(10, result)
        result = self.run_smart_contract(engine, path, 'plus', -1)
        self.assertEqual(-1, result)
        result = self.run_smart_contract(engine, path, 'plus', 0)
        self.assertEqual(0, result)

    def test_positive_builtin_type(self):
        path = self.get_contract_path('PositiveBuiltinType.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'plus', FindOptions.DESERIALIZE_VALUES)
        self.assertEqual(+FindOptions.DESERIALIZE_VALUES, result)

    # endregion

    # region StrBytesMultiplication

    def test_bytes_multiplication_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + BinaryOp.StrBytesMul.build(Type.bytes).bytecode
            + Opcode.RET
        )

        path = self.get_contract_path('BytesMultiplication.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'bytes_mult', b'a', 4,
                                         expected_result_type=bytes)
        self.assertEqual(b'aaaa', result)
        result = self.run_smart_contract(engine, path, 'bytes_mult', b'unit', 50,
                                         expected_result_type=bytes)
        self.assertEqual(b'unit' * 50, result)

    def test_bytes_multiplication_operation_augmented_assignment(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + BinaryOp.StrBytesMul.build(Type.bytes).bytecode
            + Opcode.STARG0
            + Opcode.LDARG0
            + Opcode.RET
        )

        path = self.get_contract_path('BytesMultiplicationAugmentedAssignment.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', b'unit', 50,
                                         expected_result_type=bytes)
        self.assertEqual(b'unit' * 50, result)

    def test_bytes_multiplication_builtin_type(self):
        path = self.get_contract_path('BytesMultiplicationBuiltinType.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'bytes_mult', b'unit test', FindOptions.VALUES_ONLY,
                                         expected_result_type=bytes)
        self.assertEqual(b'unit test' * FindOptions.VALUES_ONLY, result)

    def test_str_multiplication_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + BinaryOp.StrBytesMul.bytecode
            + Opcode.RET
        )

        path = self.get_contract_path('StringMultiplication.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'str_mult', 'a', 4)
        self.assertEqual('aaaa', result)
        result = self.run_smart_contract(engine, path, 'str_mult', 'unit', 50)
        self.assertEqual('unit' * 50, result)

    def test_str_multiplication_operation_augmented_assignment(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + BinaryOp.StrBytesMul.bytecode
            + Opcode.STARG0
            + Opcode.LDARG0
            + Opcode.RET
        )

        path = self.get_contract_path('StringMultiplicationAugmentedAssignment.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 'unit', 50)
        self.assertEqual('unit' * 50, result)

    def test_str_multiplication_builtin_type(self):
        path = self.get_contract_path('StringMultiplicationBuiltinType.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'str_mult', 'unit test', FindOptions.VALUES_ONLY)
        self.assertEqual('unit test' * FindOptions.VALUES_ONLY, result)

    # endregion

    # region Subtraction

    def test_subtraction_operation(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.SUB
            + Opcode.RET
        )

        path = self.get_contract_path('Subtraction.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'sub', 10, 3)
        self.assertEqual(7, result)
        result = self.run_smart_contract(engine, path, 'sub', -42, -24)
        self.assertEqual(-18, result)
        result = self.run_smart_contract(engine, path, 'sub', -42, 24)
        self.assertEqual(-66, result)

    def test_subtraction_augmented_assignment(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.SUB
            + Opcode.STARG0
            + Opcode.RET
        )

        path = self.get_contract_path('SubtractionAugmentedAssignment.py')
        output = Boa3.compile(path)

        self.assertEqual(expected_output, output)

    def test_subtraction_builtin_type(self):
        path = self.get_contract_path('SubtractionBuiltinType.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', FindOptions.DESERIALIZE_VALUES, FindOptions.VALUES_ONLY)
        self.assertEqual(FindOptions.DESERIALIZE_VALUES - FindOptions.VALUES_ONLY, result)

    # endregion



from __future__ import annotations

import ast
from typing import Dict, Optional

from boa3 import constants
from boa3.model import set_internal_call
from boa3.model.builtin.internal.internalmethod import IInternalMethod
from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.method import Method
from boa3.model.symbol import ISymbol
from boa3.model.variable import Variable


class InnerDeployMethod(IInternalMethod):

    @classmethod
    def instance(cls) -> InnerDeployMethod:
        return _INNER_DEPLOY_METHOD

    def __init__(self):
        from boa3.model.type.type import Type
        identifier = constants.DEPLOY_METHOD_ID
        args: Dict[str, Variable] = {
            'data': Variable(Type.any),
            'update': Variable(Type.bool)
        }
        super().__init__(identifier, args, return_type=Type.none)

        self.is_public = True
        self._origin_node = set_internal_call(ast.parse(self._body).body[0])

    @property
    def _body(self) -> Optional[str]:
        method_args = [f"{var_id}: {var_type.type.raw_identifier}"
                       for var_id, var_type in self.args.items()
                       ]

        return f"def {self.identifier}({', '.join(method_args)}): return"

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    def copy(self) -> InnerDeployMethod:
        return InnerDeployMethod()

    @classmethod
    def is_valid_deploy_method(cls, symbol: ISymbol) -> bool:
        if isinstance(symbol, InnerDeployMethod):
            return True
        if isinstance(symbol, IBuiltinMethod):
            return False
        if not isinstance(symbol, Method):
            return False
        else:
            if not symbol.is_public:
                return False

            reference_args = list(_INNER_DEPLOY_METHOD.args.values())
            from boa3.model.type.type import Type
            if symbol.return_type is not Type.none:
                return False

            if len(symbol.args) != len(reference_args):
                return False

            for index, arg in enumerate(symbol.args.values()):
                if not reference_args[index].type.is_equal(arg.type):
                    return False

        return True


_INNER_DEPLOY_METHOD = InnerDeployMethod()



from boa3.builtin.compile_time import public

from boa3_test.test_sc.import_test.sample_package.package import another_module, sample_module


@public
def call_imported_method() -> dict:
    a: another_module.Dict[another_module.Any, another_module.Any] = another_module.EmptyDict()
    return a


@public
def call_imported_variable() -> list:
    return sample_module.empty_list



from boa3.boa3 import Boa3
from boa3.exception import CompilerError
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo3.contracts import FindOptions
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.testengine import TestEngine


class TestLogical(BoaTest):
    default_folder: str = 'test_sc/logical_test'

    # region BoolAnd

    def test_boolean_and(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.BOOLAND
            + Opcode.RET
        )

        path = self.get_contract_path('BoolAnd.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', True, True)
        self.assertEqual(True, result)
        result = self.run_smart_contract(engine, path, 'Main', True, False)
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'Main', False, True)
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'Main', False, False)
        self.assertEqual(False, result)

    def test_mismatched_type_binary_operation(self):
        path = self.get_contract_path('MismatchedOperandAnd.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    # endregion

    # region BoolNot

    def test_boolean_not(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0
            + Opcode.NOT
            + Opcode.RET
        )

        path = self.get_contract_path('BoolNot.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', True)
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'Main', False)
        self.assertEqual(True, result)

    def test_mismatched_type_unary_operation(self):
        path = self.get_contract_path('MismatchedOperandNot.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    # endregion

    # region BoolOr

    def test_boolean_or(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.BOOLOR
            + Opcode.RET
        )

        path = self.get_contract_path('BoolOr.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', True, True)
        self.assertEqual(True, result)
        result = self.run_smart_contract(engine, path, 'Main', True, False)
        self.assertEqual(True, result)
        result = self.run_smart_contract(engine, path, 'Main', False, True)
        self.assertEqual(True, result)
        result = self.run_smart_contract(engine, path, 'Main', False, False)
        self.assertEqual(False, result)

    def test_sequence_boolean_or(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x03'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.BOOLOR
            + Opcode.LDARG2
            + Opcode.BOOLOR
            + Opcode.RET
        )

        path = self.get_contract_path('BoolOrThreeElements.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', True, False, False)
        self.assertEqual(True, result)
        result = self.run_smart_contract(engine, path, 'Main', False, True, False)
        self.assertEqual(True, result)
        result = self.run_smart_contract(engine, path, 'Main', False, False, False)
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'Main', True, True, True)
        self.assertEqual(True, result)

    # endregion

    # region LeftShift

    def test_logic_left_shift(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.SHL
            + Opcode.RET
        )

        path = self.get_contract_path('LogicLeftShift.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', int('100', 2), 2)
        self.assertEqual(int('10000', 2), result)
        result = self.run_smart_contract(engine, path, 'Main', int('11', 2), 1)
        self.assertEqual(int('110', 2), result)
        result = self.run_smart_contract(engine, path, 'Main', int('101010', 2), 4)
        self.assertEqual(int('1010100000', 2), result)

    def test_logic_left_shift_builtin_type(self):
        path = self.get_contract_path('LogicLeftShiftBuiltinType.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', FindOptions.NONE, FindOptions.DESERIALIZE_VALUES)
        self.assertEqual(FindOptions.NONE << FindOptions.DESERIALIZE_VALUES, result)

    def test_mismatched_type_logic_left_shift(self):
        path = self.get_contract_path('MismatchedOperandLogicLeftShift.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    # endregion

    # region LogicAnd

    def test_logic_and_with_bool_operand(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.AND
            + Opcode.RET
        )

        path = self.get_contract_path('LogicAndBool.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', True, True)
        self.assertEqual(True, result)
        result = self.run_smart_contract(engine, path, 'Main', True, False)
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'Main', False, False)
        self.assertEqual(False, result)

    def test_logic_and_with_int_operand(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.AND
            + Opcode.RET
        )

        path = self.get_contract_path('LogicAndInt.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 4, 6)
        self.assertEqual(4 & 6, result)
        result = self.run_smart_contract(engine, path, 'Main', 40, 6)
        self.assertEqual(40 & 6, result)
        result = self.run_smart_contract(engine, path, 'Main', -4, 32)
        self.assertEqual(-4 & 32, result)

    def test_logic_and_builtin_type(self):
        path = self.get_contract_path('LogicAndBuiltinType.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', FindOptions.NONE, FindOptions.DESERIALIZE_VALUES)
        self.assertEqual(FindOptions.NONE & FindOptions.DESERIALIZE_VALUES, result)

    def test_mismatched_type_logic_and(self):
        path = self.get_contract_path('MismatchedOperandLogicAnd.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    # endregion

    # region LogicNot

    def test_logic_not_with_bool_operand(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0
            + Opcode.INVERT
            + Opcode.RET
        )

        path = self.get_contract_path('LogicNotBool.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', True)
        self.assertEqual(-2, result)
        result = self.run_smart_contract(engine, path, 'Main', False)
        self.assertEqual(-1, result)

    def test_logic_not_builtin_type(self):
        path = self.get_contract_path('LogicNotBuiltinType.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', FindOptions.VALUES_ONLY)
        self.assertEqual(~FindOptions.VALUES_ONLY, result)

    def test_logic_not_with_int_operand(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0
            + Opcode.INVERT
            + Opcode.RET
        )

        path = self.get_contract_path('LogicNotInt.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 4)
        self.assertEqual(-5, result)
        result = self.run_smart_contract(engine, path, 'Main', 40)
        self.assertEqual(-41, result)
        result = self.run_smart_contract(engine, path, 'Main', -4)
        self.assertEqual(3, result)

    def test_mismatched_type_logic_not(self):
        path = self.get_contract_path('MismatchedOperandLogicNot.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    # endregion

    # region LogicOr

    def test_logic_or_with_bool_operand(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.OR
            + Opcode.RET
        )

        path = self.get_contract_path('LogicOrBool.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', True, True)
        self.assertEqual(True, result)
        result = self.run_smart_contract(engine, path, 'Main', True, False)
        self.assertEqual(True, result)
        result = self.run_smart_contract(engine, path, 'Main', False, False)
        self.assertEqual(False, result)

    def test_logic_or_with_int_operand(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.OR
            + Opcode.RET
        )

        path = self.get_contract_path('LogicOrInt.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 4, 6)
        self.assertEqual(4 | 6, result)
        result = self.run_smart_contract(engine, path, 'Main', 40, 6)
        self.assertEqual(40 | 6, result)
        result = self.run_smart_contract(engine, path, 'Main', -4, 32)
        self.assertEqual(-4 | 32, result)

    def test_logic_or_builtin_type(self):
        path = self.get_contract_path('LogicOrBuiltinType.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', FindOptions.REMOVE_PREFIX, FindOptions.KEYS_ONLY,
                                         expected_result_type=int)
        int_value_to_check = int(FindOptions.REMOVE_PREFIX | FindOptions.KEYS_ONLY)
        self.assertEqual(int_value_to_check, result)

        result = self.run_smart_contract(engine, path, 'main', FindOptions.REMOVE_PREFIX, FindOptions.KEYS_ONLY,
                                         expected_result_type=FindOptions)
        self.assertEqual(FindOptions.REMOVE_PREFIX | FindOptions.KEYS_ONLY, result)

        result = self.run_smart_contract(engine, path, 'main', 2, 4,
                                         expected_result_type=int)
        int_value_to_check = int(2 | 4)
        self.assertEqual(int_value_to_check, result)

        result = self.run_smart_contract(engine, path, 'main', 0, 123456789,
                                         expected_result_type=int)
        self.assertEqual(123456789, result)

        result = self.run_smart_contract(engine, path, 'main', 123456789, 0,
                                         expected_result_type=int)
        self.assertEqual(123456789, result)

    def test_mismatched_type_logic_or(self):
        path = self.get_contract_path('MismatchedOperandLogicOr.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    # endregion

    # region LogicXor

    def test_logic_xor_with_bool_operand(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.XOR
            + Opcode.RET
        )

        path = self.get_contract_path('LogicXorBool.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', True, True)
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'Main', True, False)
        self.assertEqual(True, result)
        result = self.run_smart_contract(engine, path, 'Main', False, False)
        self.assertEqual(False, result)

    def test_logic_xor_with_int_operand(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.XOR
            + Opcode.RET
        )

        path = self.get_contract_path('LogicXorInt.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 4, 6)
        self.assertEqual(4 ^ 6, result)
        result = self.run_smart_contract(engine, path, 'Main', 40, 6)
        self.assertEqual(40 ^ 6, result)
        result = self.run_smart_contract(engine, path, 'Main', -4, 32)
        self.assertEqual(-4 ^ 32, result)

    def test_logic_xor_builtin_type(self):
        path = self.get_contract_path('LogicXorBuiltinType.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', FindOptions.NONE, FindOptions.DESERIALIZE_VALUES)
        self.assertEqual(FindOptions.NONE ^ FindOptions.DESERIALIZE_VALUES, result)

    def test_mismatched_type_logic_xor(self):
        path = self.get_contract_path('MismatchedOperandLogicXor.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    # endregion

    # region Mixed

    def test_logic_augmented_assignment(self):
        path = self.get_contract_path('AugmentedAssignmentOperators.py')
        engine = TestEngine()

        a = 1
        b = 4
        result = self.run_smart_contract(engine, path, 'right_shift', a, b)
        a >>= b
        expected_result = a
        self.assertEqual(expected_result, result)

        a = 4
        b = 1
        result = self.run_smart_contract(engine, path, 'left_shift', a, b)
        a <<= b
        expected_result = a
        self.assertEqual(expected_result, result)

        a = 255
        b = 123
        result = self.run_smart_contract(engine, path, 'l_and', a, b)
        a &= b
        expected_result = a
        self.assertEqual(expected_result, result)

        a = 255
        b = 123
        result = self.run_smart_contract(engine, path, 'l_or', a, b)
        a |= b
        expected_result = a
        self.assertEqual(expected_result, result)

        a = 255
        b = 123
        result = self.run_smart_contract(engine, path, 'xor', a, b)
        a ^= b
        expected_result = a
        self.assertEqual(expected_result, result)

    def test_boa2_logic_test(self):
        path = self.get_contract_path('BinOpBoa2Test.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main', '&', 4, 4)
        self.assertEqual(4, result)

        result = self.run_smart_contract(engine, path, 'main', '|', 4, 3)
        self.assertEqual(7, result)

        result = self.run_smart_contract(engine, path, 'main', '|', 4, 8)
        self.assertEqual(12, result)

        result = self.run_smart_contract(engine, path, 'main', '^', 4, 4)
        self.assertEqual(0, result)

        result = self.run_smart_contract(engine, path, 'main', '^', 4, 2)
        self.assertEqual(6, result)

        result = self.run_smart_contract(engine, path, 'main', '>>', 16, 2)
        self.assertEqual(4, result)

        result = self.run_smart_contract(engine, path, 'main', '>>', 16, 0)
        self.assertEqual(16, result)

        result = self.run_smart_contract(engine, path, 'main', '%', 16, 2)
        self.assertEqual(0, result)

        result = self.run_smart_contract(engine, path, 'main', '%', 16, 11)
        self.assertEqual(5, result)

        result = self.run_smart_contract(engine, path, 'main', '//', 16, 2)
        self.assertEqual(8, result)

        result = self.run_smart_contract(engine, path, 'main', '//', 16, 7)
        self.assertEqual(2, result)

        result = self.run_smart_contract(engine, path, 'main', '~', 16, 0)
        self.assertEqual(-17, result)

        result = self.run_smart_contract(engine, path, 'main', '~', -3, 0)
        self.assertEqual(2, result)

    def test_mixed_operations(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x03'
            + Opcode.LDARG0
            + Opcode.NOT
            + Opcode.LDARG1
            + Opcode.LDARG2
            + Opcode.BOOLOR
            + Opcode.BOOLAND
            + Opcode.RET
        )

        path = self.get_contract_path('MixedOperations.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', True, False, False)
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'Main', False, True, False)
        self.assertEqual(True, result)
        result = self.run_smart_contract(engine, path, 'Main', False, False, False)
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'Main', True, True, True)
        self.assertEqual(False, result)

    def test_logic_operation_with_return_and_stack_filled(self):
        path = self.get_contract_path('LogicOperationWithReturnAndStackFilled.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(True, result)

    # endregion

    # region RightShift

    def test_logic_right_shift(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG0
            + Opcode.LDARG1
            + Opcode.SHR
            + Opcode.RET
        )

        path = self.get_contract_path('LogicRightShift.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', int('10000', 2), 2)
        self.assertEqual(int('100', 2), result)
        result = self.run_smart_contract(engine, path, 'Main', int('110', 2), 1)
        self.assertEqual(int('11', 2), result)
        result = self.run_smart_contract(engine, path, 'Main', int('1010100000', 2), 4)
        self.assertEqual(int('101010', 2), result)

    def test_logic_right_shift_builtin_type(self):
        path = self.get_contract_path('LogicRightShiftBuiltinType.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', FindOptions.NONE, FindOptions.DESERIALIZE_VALUES)
        self.assertEqual(FindOptions.NONE >> FindOptions.DESERIALIZE_VALUES, result)

    def test_mismatched_type_logic_right_shift(self):
        path = self.get_contract_path('MismatchedOperandLogicRightShift.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    # endregion



from typing import Any, Optional, Sized

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.type.classes.classtype import ClassType
from boa3.model.type.itype import IType


class SuperMethod(IBuiltinMethod):

    def __init__(self, value_type: IType = None):
        if isinstance(value_type, ClassType) and len(value_type.bases) > 0:
            # TODO: change when inheritance with multiple bases is implemented
            return_type = value_type.bases[0]
        else:
            from boa3.model.type.type import Type
            return_type = Type.any

        identifier = 'super'
        self._super_type = value_type if isinstance(value_type, IType) else None
        super().__init__(identifier, return_type=return_type)

    @property
    def identifier(self) -> str:
        if self._super_type is None:
            return self._identifier
        return '-{0}_from_{1}'.format(self._identifier, self._super_type._identifier)

    @property
    def is_supported(self) -> bool:
        return self._super_type is not None

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return

    def build(self, value: Any) -> IBuiltinMethod:
        if isinstance(value, Sized) and len(value) > 0:
            value = value[0]
        if value == self._super_type:
            return self
        if isinstance(value, IType):
            return SuperMethod(value)
        return super().build(value)



from boa3.boa3 import Boa3
from boa3.exception import CompilerWarning
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.testengine import TestEngine


class TestIf(BoaTest):
    default_folder: str = 'test_sc/if_test'

    def test_if_constant_condition(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x00'
            + Opcode.PUSH0      # a = 0
            + Opcode.STLOC0
            + Opcode.PUSHT
            + Opcode.JMPIFNOT   # if True
            + Integer(4).to_byte_array(min_length=1, signed=True)
            + Opcode.PUSH2     # a = a + 2
            + Opcode.STLOC0
            + Opcode.LDLOC0     # return a
            + Opcode.RET
        )

        path = self.get_contract_path('ConstantCondition.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(2, result)

    def test_if_variable_condition(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x01'
            + Opcode.PUSH0      # a = 0
            + Opcode.STLOC0
            + Opcode.LDARG0
            + Opcode.JMPIFNOT   # if arg0
            + Integer(4).to_byte_array(min_length=1, signed=True)
            + Opcode.PUSH2      # a = a + 2
            + Opcode.STLOC0
            + Opcode.LDLOC0     # return a
            + Opcode.RET
        )

        path = self.get_contract_path('VariableCondition.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', True)
        self.assertEqual(2, result)
        result = self.run_smart_contract(engine, path, 'Main', False)
        self.assertEqual(0, result)

    def test_if_no_condition(self):
        path = self.get_contract_path('IfWithoutCondition.py')

        with self.assertRaises(SyntaxError):
            Boa3.compile(path)

    def test_if_no_body(self):
        path = self.get_contract_path('IfWithoutBody.py')

        with self.assertRaises(SyntaxError):
            Boa3.compile(path)

    def test_nested_if(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x02'
            + b'\x02'
            + Opcode.PUSH0      # c = 0
            + Opcode.STLOC0
            + Opcode.PUSH0      # d = c
            + Opcode.STLOC1
            + Opcode.LDARG0
            + Opcode.JMPIFNOT   # if arg0
            + Integer(13).to_byte_array(min_length=1, signed=True)
            + Opcode.PUSH2      # c = c + 2
            + Opcode.STLOC0
            + Opcode.LDARG1
            + Opcode.JMPIFNOT   # if arg1
            + Integer(4).to_byte_array(min_length=1, signed=True)
            + Opcode.PUSH3      # d = d + 3
            + Opcode.STLOC1
            + Opcode.PUSH2      # c = c + d
            + Opcode.LDLOC1
            + Opcode.ADD
            + Opcode.STLOC0
            + Opcode.LDLOC0     # return c
            + Opcode.RET
        )

        path = self.get_contract_path('NestedIf.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', True, True)
        self.assertEqual(5, result)
        result = self.run_smart_contract(engine, path, 'Main', True, False)
        self.assertEqual(2, result)
        result = self.run_smart_contract(engine, path, 'Main', False, True)
        self.assertEqual(0, result)
        result = self.run_smart_contract(engine, path, 'Main', False, False)
        self.assertEqual(0, result)

    def test_if_else(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x01'
            + Opcode.PUSH0      # a = 0
            + Opcode.STLOC0
            + Opcode.LDARG0
            + Opcode.JMPIFNOT   # if arg0
            + Integer(6).to_byte_array(min_length=1, signed=True)
            + Opcode.PUSH2      # a = a + 2
            + Opcode.STLOC0
            + Opcode.JMP        # else
            + Integer(4).to_byte_array(min_length=1, signed=True)
            + Opcode.PUSH10     # a = 10
            + Opcode.STLOC0
            + Opcode.LDLOC0     # return a
            + Opcode.RET
        )
        path = self.get_contract_path('IfElse.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', True)
        self.assertEqual(2, result)
        result = self.run_smart_contract(engine, path, 'Main', False)
        self.assertEqual(10, result)

    def test_else_no_body(self):
        path = self.get_contract_path('ElseWithoutBody.py')

        with self.assertRaises(SyntaxError):
            Boa3.compile(path)

    def test_if_elif(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x01'
            + Opcode.PUSH0      # a = 0
            + Opcode.STLOC0
            + Opcode.LDARG0
            + Opcode.JMPIFNOT   # if arg0
            + Integer(6).to_byte_array(min_length=1)
            + Opcode.PUSH2      # a = a + 2
            + Opcode.STLOC0
            + Opcode.JMP
            + Integer(7).to_byte_array(min_length=1)
            + Opcode.LDARG0
            + Opcode.JMPIFNOT   # elif arg0
            + Integer(4).to_byte_array(min_length=1)
            + Opcode.PUSH10     # a = 10
            + Opcode.STLOC0
            + Opcode.LDLOC0     # return a
            + Opcode.RET
        )

        path = self.get_contract_path('IfElif.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', True)
        self.assertEqual(2, result)
        result = self.run_smart_contract(engine, path, 'Main', False)
        self.assertEqual(0, result)

    def test_elif_no_condition(self):
        path = self.get_contract_path('ElifWithoutCondition.py')

        with self.assertRaises(SyntaxError):
            output = Boa3.compile(path)

    def test_elif_no_body(self):
        path = self.get_contract_path('ElifWithoutBody.py')

        with self.assertRaises(SyntaxError):
            output = Boa3.compile(path)

    def test_if_relational_condition(self):
        jmp_address = Integer(4).to_byte_array(min_length=1, signed=True)

        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x01'
            + Opcode.PUSH0      # a = 0
            + Opcode.STLOC0
            + Opcode.LDARG0
            + Opcode.PUSH10
            + Opcode.LT
            + Opcode.JMPIFNOT   # if c < 10
            + jmp_address
            + Opcode.PUSH2      # a = a + 2
            + Opcode.STLOC0
            + Opcode.LDLOC0     # return a
            + Opcode.RET
        )

        path = self.get_contract_path('RelationalCondition.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 5)
        self.assertEqual(2, result)
        result = self.run_smart_contract(engine, path, 'Main', 10)
        self.assertEqual(0, result)

    def test_if_multiple_branches(self):
        twenty = Integer(20).to_byte_array()
        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x01'
            + Opcode.LDARG0
            + Opcode.PUSH0
            + Opcode.LT
            + Opcode.JMPIFNOT       # if arg0 < 0
            + Integer(6).to_byte_array(min_length=1)
            + Opcode.PUSH0              # a = 0
            + Opcode.STLOC0
            + Opcode.JMP
            + Integer(32).to_byte_array(min_length=1)
            + Opcode.LDARG0
            + Opcode.PUSH5
            + Opcode.LT
            + Opcode.JMPIFNOT       # elif arg0 < 5
            + Integer(6).to_byte_array(min_length=1)
            + Opcode.PUSH5              # a = 5
            + Opcode.STLOC0
            + Opcode.JMP
            + Integer(23).to_byte_array(min_length=1)
            + Opcode.LDARG0
            + Opcode.PUSH10
            + Opcode.LT
            + Opcode.JMPIFNOT       # elif arg0 < 10
            + Integer(6).to_byte_array(min_length=1)
            + Opcode.PUSH10             # a = 10
            + Opcode.STLOC0
            + Opcode.JMP
            + Integer(14).to_byte_array(min_length=1)
            + Opcode.LDARG0
            + Opcode.PUSH15
            + Opcode.LT
            + Opcode.JMPIFNOT       # elif arg0 < 15
            + Integer(6).to_byte_array(min_length=1)
            + Opcode.PUSH15             # a = 15
            + Opcode.STLOC0
            + Opcode.JMP            # else
            + Integer(5).to_byte_array(min_length=1)
            + Opcode.PUSHINT8 + twenty  # a = 20
            + Opcode.STLOC0
            + Opcode.LDLOC0     # return a
            + Opcode.RET
        )

        path = self.get_contract_path('MultipleBranches.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', -10)
        self.assertEqual(0, result)
        result = self.run_smart_contract(engine, path, 'Main', 2)
        self.assertEqual(5, result)
        result = self.run_smart_contract(engine, path, 'Main', 7)
        self.assertEqual(10, result)
        result = self.run_smart_contract(engine, path, 'Main', 13)
        self.assertEqual(15, result)
        result = self.run_smart_contract(engine, path, 'Main', 17)
        self.assertEqual(20, result)
        result = self.run_smart_contract(engine, path, 'Main', 23)
        self.assertEqual(20, result)

    def test_if_expression_variable_condition(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x01'
            + Opcode.LDARG0
            + Opcode.JMPIFNOT   # a = 2 if arg0 else 3
            + Integer(5).to_byte_array(min_length=1, signed=True)
            + Opcode.PUSH2      # 2
            + Opcode.JMP        # else
            + Integer(3).to_byte_array(min_length=1, signed=True)
            + Opcode.PUSH3      # 3
            + Opcode.STLOC0
            + Opcode.LDLOC0     # return a
            + Opcode.RET
        )
        path = self.get_contract_path('IfExpVariableCondition.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', True)
        self.assertEqual(2, result)
        result = self.run_smart_contract(engine, path, 'Main', False)
        self.assertEqual(3, result)

    def test_if_expression_without_else_branch(self):
        path = self.get_contract_path('IfExpWithoutElse.py')

        with self.assertRaises(SyntaxError):
            output = Boa3.compile(path)

    def test_if_expression_mismatched_types(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x01'
            + Opcode.LDARG0
            + Opcode.JMPIFNOT   # a = 2 if condition else None
            + Integer(5).to_byte_array(min_length=1, signed=True)
            + Opcode.PUSH2      # 2
            + Opcode.JMP        # else
            + Integer(3).to_byte_array(min_length=1, signed=True)
            + Opcode.PUSHNULL   # None
            + Opcode.STLOC0
            + Opcode.LDLOC0     # return a
            + Opcode.RET
        )

        path = self.get_contract_path('MismatchedIfExp.py')
        output = self.assertCompilerLogs(CompilerWarning.TypeCasting, path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', True)
        self.assertEqual(2, result)
        result = self.run_smart_contract(engine, path, 'Main', False)
        self.assertEqual(None, result)

    def test_inner_if_else(self):
        path = self.get_contract_path('InnerIfElse.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', 4, 3, 2, 1)
        self.assertEqual(3, result)

        result = self.run_smart_contract(engine, path, 'main', 4, 3, 1, 2)
        self.assertEqual(8, result)

        result = self.run_smart_contract(engine, path, 'main', 4, 1, 2, 3)
        self.assertEqual(10, result)

        result = self.run_smart_contract(engine, path, 'main', 1, 2, 4, 3)
        self.assertEqual(1, result)

        result = self.run_smart_contract(engine, path, 'main', 1, 2, 3, 4)
        self.assertEqual(11, result)

        result = self.run_smart_contract(engine, path, 'main', 1, 3, 2, 4)
        self.assertEqual(22, result)

    def test_if_is_instance_condition(self):
        path = self.get_contract_path('IfIsInstanceCondition.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'example', 4)
        self.assertEqual(4, result)

        result = self.run_smart_contract(engine, path, 'example', '123')
        self.assertEqual(-1, result)

        result = self.run_smart_contract(engine, path, 'example', -70)
        self.assertEqual(-70, result)

        result = self.run_smart_contract(engine, path, 'example', True)
        self.assertEqual(-1, result)

    def test_if_else_is_instance_condition(self):
        path = self.get_contract_path('IfElseIsInstanceCondition.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'example', 4)
        self.assertEqual(4, result)

        result = self.run_smart_contract(engine, path, 'example', '123')
        self.assertEqual(-1, result)

        result = self.run_smart_contract(engine, path, 'example', -70)
        self.assertEqual(-70, result)

        result = self.run_smart_contract(engine, path, 'example', True)
        self.assertEqual(-1, result)

    def test_if_else_is_instance_condition_with_union_variable(self):
        path = self.get_contract_path('IfElseIsInstanceConditionWithUnionVariable.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'example', 4,
                                         expected_result_type=bytes)
        self.assertEqual(b'\x04', result)

        result = self.run_smart_contract(engine, path, 'example', '123',
                                         expected_result_type=bytes)
        self.assertEqual(b'123', result)

        result = self.run_smart_contract(engine, path, 'example', -70,
                                         expected_result_type=bytes)
        self.assertEqual(Integer(-70).to_byte_array(), result)

        result = self.run_smart_contract(engine, path, 'example', True,
                                         expected_result_type=bytes)
        self.assertEqual(b'\x01', result)

    def test_if_else_multiple_is_instance_condition_with_union_variable(self):
        path = self.get_contract_path('IfElseMultipleIsInstanceConditionWithUnionVariable.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'example', 4)
        self.assertEqual(16, result)

        result = self.run_smart_contract(engine, path, 'example', [b'123456', b'789'])
        self.assertEqual(6, result)

        result = self.run_smart_contract(engine, path, 'example', -70)
        self.assertEqual(4900, result)

        result = self.run_smart_contract(engine, path, 'example', [])
        self.assertEqual(1, result)

        result = self.run_smart_contract(engine, path, 'example', b'True')
        self.assertEqual(4, result)

    def test_variable_in_if_scopes(self):
        path = self.get_contract_path('VariablesInIfScopes.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', 1)
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, path, 'main', 2)
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, path, 'main', 3)
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, path, 'main', 4)
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, path, 'main', 5)
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, path, 'main', 6)
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, path, 'main', 7)
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, path, 'main', 8)
        self.assertEqual(False, result)

    def test_boa2_compare_test0int(self):
        path = self.get_contract_path('CompareBoa2Test0int.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', 2, 4)
        self.assertEqual(2, result)

        result = self.run_smart_contract(engine, path, 'main', 4, 2)
        self.assertEqual(3, result)

        result = self.run_smart_contract(engine, path, 'main', 2, 2)
        self.assertEqual(2, result)

    def test_boa2_compare_test0str(self):
        path = self.get_contract_path('CompareBoa2Test0str.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', 'b', 'a')
        self.assertEqual(3, result)

        result = self.run_smart_contract(engine, path, 'main', 'a', 'b')
        self.assertEqual(2, result)

    def test_boa2_compare_test1(self):
        path = self.get_contract_path('CompareBoa2Test1.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main', 1, 2, 3, 4)
        self.assertEqual(11, result)

        result = self.run_smart_contract(engine, path, 'main', 1, 2, 4, 3)
        self.assertEqual(1, result)

        result = self.run_smart_contract(engine, path, 'main', 1, 4, 3, 5)
        self.assertEqual(22, result)

        result = self.run_smart_contract(engine, path, 'main', 4, 1, 5, 3)
        self.assertEqual(3, result)

        result = self.run_smart_contract(engine, path, 'main', 9, 1, 3, 5)
        self.assertEqual(10, result)

        result = self.run_smart_contract(engine, path, 'main', 9, 5, 3, 5)
        self.assertEqual(8, result)

    def test_boa2_compare_test2(self):
        path = self.get_contract_path('CompareBoa2Test2.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main', 2, 2)
        self.assertEqual(True, result)

        result = self.run_smart_contract(engine, path, 'main', 2, 3)
        self.assertEqual(False, result)

    def test_boa2_op_call_test(self):
        path = self.get_contract_path('OpCallBoa2Test.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', 'omin', 4, 4)
        self.assertEqual(4, result)

        result = self.run_smart_contract(engine, path, 'main', 'omin', -4, 4)
        self.assertEqual(-4, result)

        result = self.run_smart_contract(engine, path, 'main', 'omin', 16, 0)
        self.assertEqual(0, result)

        result = self.run_smart_contract(engine, path, 'main', 'omax', 4, 4)
        self.assertEqual(4, result)

        result = self.run_smart_contract(engine, path, 'main', 'omax', -4, 4)
        self.assertEqual(4, result)

        result = self.run_smart_contract(engine, path, 'main', 'omax', 16, 0)
        self.assertEqual(16, result)

        from boa3.neo.cryptography import sha256, hash160
        from boa3.neo.vm.type.String import String
        result = self.run_smart_contract(engine, path, 'main', 'sha256', 'abc', 4)
        self.assertEqual(sha256(String('abc').to_bytes()), result)

        result = self.run_smart_contract(engine, path, 'main', 'hash160', 'abc', 4)
        self.assertEqual(hash160(String('abc').to_bytes()), result)

    def test_boa2_test_many_elif(self):
        path = self.get_contract_path('TestManyElifBoa2.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', 1)
        self.assertEqual(2, result)

        result = self.run_smart_contract(engine, path, 'main', 3)
        self.assertEqual(4, result)

        result = self.run_smart_contract(engine, path, 'main', 16)
        self.assertEqual(17, result)

        result = self.run_smart_contract(engine, path, 'main', 22)
        self.assertEqual(-1, result)

    def test_if_with_inner_while(self):
        path = self.get_contract_path('IfWithInnerWhile.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'Main', True)
        self.assertEqual('{[]}', result)

        result = self.run_smart_contract(engine, path, 'Main', False)
        self.assertEqual('{[value1,value2,value3]}', result)

    def test_if_with_inner_for(self):
        path = self.get_contract_path('IfWithInnerFor.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'Main', True)
        self.assertEqual('{[]}', result)

        result = self.run_smart_contract(engine, path, 'Main', False)
        self.assertEqual('{[value1,value2,value3]}', result)

    def test_if_implicit_boolean(self):
        path = self.get_contract_path('IfImplicitBoolean.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', 1)
        self.assertEqual(True, result)

        result = self.run_smart_contract(engine, path, 'main', 0)
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, path, 'main', 'unit_test')
        self.assertEqual(True, result)

        result = self.run_smart_contract(engine, path, 'main', '')
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, path, 'main', b'unit test')
        self.assertEqual(True, result)

        result = self.run_smart_contract(engine, path, 'main', b'')
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, path, 'main', [1, 2, 3, 4])
        self.assertEqual(True, result)

        result = self.run_smart_contract(engine, path, 'main', [])
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, path, 'main', {'a': 1, 'b': 2})
        self.assertEqual(True, result)

        result = self.run_smart_contract(engine, path, 'main', {})
        self.assertEqual(False, result)

    def test_if_implicit_boolean_literal(self):
        path = self.get_contract_path('IfImplicitBooleanLiteral.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(4, result)

    def test_if_pass(self):
        path = self.get_contract_path('IfPass.py')
        engine = TestEngine()

        output = Boa3.compile(path)
        self.assertIn(Opcode.NOP, output)

        result = self.run_smart_contract(engine, path, 'main', True)
        self.assertEqual(result, 0)

    def test_else_pass(self):
        path = self.get_contract_path('ElsePass.py')
        engine = TestEngine()

        output = Boa3.compile(path)
        self.assertIn(Opcode.NOP, output)

        result = self.run_smart_contract(engine, path, 'main', True)
        self.assertEqual(result, 5)
        result = self.run_smart_contract(engine, path, 'main', False)
        self.assertEqual(result, 0)

    def test_if_else_pass(self):
        path = self.get_contract_path('IfElsePass.py')
        engine = TestEngine()

        output = Boa3.compile(path)
        n_nop = 0
        for byte_value in output:
            if byte_value == int.from_bytes(Opcode.NOP.value, 'little'):
                n_nop += 1
        self.assertEqual(n_nop, 2)

        result = self.run_smart_contract(engine, path, 'main', True)
        self.assertEqual(result, 0)

    def test_if_is_none(self):
        path = self.get_contract_path('IfIsNone.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', 123)
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'main', None)
        self.assertEqual(True, result)

    def test_if_is_not_none(self):
        path = self.get_contract_path('IfIsNotNone.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', 123)
        self.assertEqual(True, result)
        result = self.run_smart_contract(engine, path, 'main', None)
        self.assertEqual(False, result)

    def test_if_is_none_type_check(self):
        path = self.get_contract_path('IfIsNoneTypeCheck.py')
        engine = TestEngine()

        test_input = bytes(20)
        result = self.run_smart_contract(engine, path, 'main', test_input,
                                         expected_result_type=bytes)
        self.assertEqual(test_input, result)

        result = self.run_smart_contract(engine, path, 'main', None)
        self.assertEqual(None, result)



from boa3.boa3 import Boa3
from boa3.exception import CompilerError
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.TestExecutionException import TestExecutionException
from boa3_test.tests.test_classes.testengine import TestEngine


class TestImport(BoaTest):
    default_folder: str = 'test_sc/import_test'

    def test_import_typing(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.NEWARRAY0
            + Opcode.STLOC0
            + Opcode.RET        # return
        )
        path = self.get_contract_path('ImportTyping.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_import_typing_with_alias(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.NEWARRAY0
            + Opcode.STLOC0
            + Opcode.RET        # return
        )
        path = self.get_contract_path('ImportTypingWithAlias.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_import_user_module(self):
        expected_output = (
            Opcode.CALL
            + Integer(3).to_byte_array(min_length=1, signed=True)
            + Opcode.RET
            + Opcode.NEWARRAY0  # imported module's function
            + Opcode.RET  # return
            + Opcode.INITSSLOT + b'\x01'
            + Opcode.NEWARRAY0  # imported module's variable in the init
            + Opcode.STSFLD0
            + Opcode.RET
        )

        path = self.get_contract_path('ImportUserModule.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual([], result)

    def test_import_user_module_with_alias(self):
        expected_output = (
            Opcode.CALL
            + Integer(3).to_byte_array(min_length=1, signed=True)
            + Opcode.RET
            + Opcode.NEWARRAY0  # imported module's function
            + Opcode.RET  # return
            + Opcode.INITSSLOT + b'\x01'
            + Opcode.NEWARRAY0  # imported module's variable in the init
            + Opcode.STSFLD0
            + Opcode.RET
        )

        path = self.get_contract_path('ImportUserModuleWithAlias.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual([], result)

    def test_import_user_module_with_global_variables(self):
        expected_output = (
            Opcode.LDSFLD0  # b = a
            + Opcode.RET
            + Opcode.NEWARRAY0  # imported function
            + Opcode.RET  # return
            + Opcode.INITSSLOT + b'\x02'
            + Opcode.CALL  # a = UserModule.EmptyList()
            + Integer(-4).to_byte_array(min_length=1, signed=True)
            + Opcode.STSFLD0
            + Opcode.NEWARRAY0     # module variable empty_list from import
            + Opcode.STSFLD1
            + Opcode.RET
        )

        path = self.get_contract_path('FromImportWithGlobalVariables.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual([], result)

    def test_import_variable(self):
        expected_output = (
            Opcode.LDSFLD0  # return empty_list
            + Opcode.RET
            + Opcode.NEWARRAY0  # imported function
            + Opcode.RET  # return
            + Opcode.INITSSLOT + b'\x01'
            + Opcode.NEWARRAY0  # imported variable
            + Opcode.STSFLD0
            + Opcode.RET
        )

        path = self.get_contract_path('FromImportVariable.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual([], result)

    def test_variable_from_imported_module(self):
        path = self.get_contract_path('variable_import', 'VariableFromImportedModule.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'get_foo', expected_result_type=bytes)
        self.assertEqual(b'Foo', result)

        result = self.run_smart_contract(engine, path, 'get_bar')
        self.assertEqual('bar', result)

    def test_variable_access_from_imported_module(self):
        path = self.get_contract_path('variable_import', 'VariableAccessFromImportedModule.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'get_foo', expected_result_type=bytes)
        self.assertEqual(b'Foo', result)

        result = self.run_smart_contract(engine, path, 'get_bar')
        self.assertEqual('bar', result)

    def test_typing_python_library(self):
        path = self.get_contract_path('ImportPythonLib.py')
        self.assertCompilerLogs(CompilerError.UnresolvedReference, path)

    def test_from_typing_import(self):
        expected_output = (
            Opcode.NEWARRAY0
            + Opcode.RET        # return
            + Opcode.INITSSLOT + b'\x01'
            + Opcode.NEWARRAY0
            + Opcode.STSFLD0
            + Opcode.RET
        )

        path = self.get_contract_path('FromImportTyping.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_from_typing_import_with_alias(self):
        expected_output = (
            Opcode.NEWARRAY0
            + Opcode.RET        # return
        )

        path = self.get_contract_path('FromImportTypingWithAlias.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_from_typing_import_not_supported_type(self):
        path = self.get_contract_path('FromImportTypingNotImplementedType.py')
        self.assertCompilerLogs(CompilerError.UnresolvedReference, path)

    def test_from_import_all(self):
        expected_output = (
            Opcode.INITSLOT  # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.CALL
            + Integer(7).to_byte_array(min_length=1, signed=True)
            + Opcode.STLOC0
            + Opcode.LDLOC0
            + Opcode.RET
            + Opcode.LDSFLD0  # return empty_list
            + Opcode.RET
            + Opcode.NEWARRAY0  # imported function
            + Opcode.RET  # return
            + Opcode.INITSSLOT + b'\x01'
            + Opcode.NEWARRAY0     # module variable empty_list from import
            + Opcode.STSFLD0
            + Opcode.RET
        )

        path = self.get_contract_path('FromImportAll.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'call_imported_method')
        self.assertEqual([], result)

        result = self.run_smart_contract(engine, path, 'call_imported_variable')
        self.assertEqual([], result)

    def test_from_import_user_module(self):
        expected_output = (
            Opcode.INITSLOT  # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.CALL
            + Integer(5).to_byte_array(min_length=1, signed=True)
            + Opcode.STLOC0
            + Opcode.LDLOC0
            + Opcode.RET
            + Opcode.NEWARRAY0  # imported function
            + Opcode.RET  # return
            + Opcode.INITSSLOT + b'\x01'
            + Opcode.NEWARRAY0     # module variable empty_list from import
            + Opcode.STSFLD0
            + Opcode.RET
        )

        path = self.get_contract_path('FromImportUserModule.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual([], result)

    def test_from_import_user_module_with_alias(self):
        expected_output = (
            Opcode.INITSLOT  # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.CALL
            + Integer(5).to_byte_array(min_length=1, signed=True)
            + Opcode.STLOC0
            + Opcode.LDLOC0
            + Opcode.RET
            + Opcode.NEWARRAY0  # imported function
            + Opcode.RET  # return
            + Opcode.INITSSLOT + b'\x01'
            + Opcode.NEWARRAY0     # module variable empty_list from import
            + Opcode.STSFLD0
            + Opcode.RET
        )

        path = self.get_contract_path('FromImportUserModuleWithAlias.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual([], result)

    def test_import_non_existent_package(self):
        path = self.get_contract_path('ImportNonExistentPackage.py')
        self.assertCompilerLogs(CompilerError.UnresolvedReference, path)

    def test_import_interop_with_alias(self):
        path = self.get_contract_path('ImportInteropWithAlias.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'Main')
        self.assertIsVoid(result)

    def test_import_user_module_recursive_import(self):
        path = self.get_contract_path('ImportUserModuleRecursiveImport.py')
        self.assertCompilerLogs(CompilerError.CircularImport, path)

    def test_from_import_user_module_recursive_import(self):
        path = self.get_contract_path('FromImportUserModuleRecursiveImport.py')
        self.assertCompilerLogs(CompilerError.CircularImport, path)

    def test_import_user_module_with_not_imported_symbols(self):
        path = self.get_contract_path('ImportUserModuleWithNotImportedSymbols.py')

        engine = TestEngine()
        self.run_smart_contract(engine, path, 'main', [], b'00000000000000000000')
        script = engine.executed_script_hash.to_array()

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main', [], script)
        self.assertEqual([], result)

        result = self.run_smart_contract(engine, path, 'main', [1, 2, 3], script)
        expected_result = []
        for x in [1, 2, 3]:
            expected_result.append([script,
                                    'notify',
                                    [x]])
        self.assertEqual(expected_result, result)

        result = self.run_smart_contract(engine, path, 'main', [1, 2, 3], b'\x01' * 20)
        self.assertEqual([], result)

        # 'with_param' is a public method, so it should be included in the manifest when imported
        result = self.run_smart_contract(engine, path, 'with_param', [1, 2, 3], b'\x01' * 20)
        self.assertEqual([], result)

        # 'without_param' is a public method, but it isn't imported, so it shouldn't be included in the manifest
        with self.assertRaisesRegex(TestExecutionException, f'{self.CANT_FIND_METHOD_MSG_PREFIX} : without_param'):
            self.run_smart_contract(engine, path, 'without_param', [1, 2, 3])

    def test_import_user_module_with_not_imported_variables(self):
        expected_output = (
            Opcode.CALL
            + Integer(5).to_byte_array(min_length=1, signed=True)
            + Opcode.RET
            + Opcode.LDSFLD1    # module function from import
            + Opcode.RET  # return
            + Opcode.LDSFLD2    # imported function
            + Opcode.RET  # return
            + Opcode.INITSSLOT + b'\x03'
            + Opcode.PUSH15     # module variable a
            + Opcode.STSFLD0
            + Opcode.PUSH10     # module variable a from import
            + Opcode.STSFLD1
            + Opcode.PUSH5      # module variable b from import
            + Opcode.STSFLD2
            + Opcode.RET
        )

        path = self.get_contract_path('ImportUserModuleWithNotImportedVariables.py')
        output, manifest = self.compile_and_save(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(5, result)

    def test_not_imported_builtin_public(self):
        path = self.get_contract_path('NotImportedBuiltinPublic.py')
        self.assertCompilerLogs(CompilerError.UnresolvedReference, path)

    def test_not_imported_builtin_from_typing(self):
        path = self.get_contract_path('NotImportedBuiltinFromTypingInReturn.py')
        self.assertCompilerLogs(CompilerError.UnresolvedReference, path)

        path = self.get_contract_path('NotImportedBuiltinFromTypingInArgs.py')
        self.assertCompilerLogs(CompilerError.UnresolvedReference, path)

        path = self.get_contract_path('NotImportedBuiltinFromTypingInSubscript.py')
        self.assertCompilerLogs(CompilerError.UnresolvedReference, path)

        path = self.get_contract_path('NotImportedBuiltinFromTypingInVariable.py')
        self.assertCompilerLogs(CompilerError.UnresolvedReference, path)

    def test_incorrect_circular_import(self):
        path = self.get_contract_path('incorrect_circular_import', 'IncorrectCircularImportDetection.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(3, result)

    def test_import_module_with_init(self):
        path = self.get_contract_path('ImportModuleWithInit.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'call_imported_method')
        self.assertEqual([], result)

        result = self.run_smart_contract(engine, path, 'call_imported_variable')
        self.assertEqual(42, result)

    def test_import_module_without_init(self):
        path = self.get_contract_path('ImportModuleWithoutInit.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'call_imported_method')
        self.assertEqual({}, result)

        result = self.run_smart_contract(engine, path, 'call_imported_variable')
        self.assertEqual([], result)

    def test_import_user_class_inner_files(self):
        inner_path = self.get_contract_path('class_import', 'ImportUserClass.py')
        path = self.get_contract_path('ImportUserClassInnerFiles.py')

        engine = TestEngine()

        result = self.run_smart_contract(engine, inner_path, 'build_example_object')
        self.assertEqual([42, '42'], result)

        result = self.run_smart_contract(engine, path, 'build_example_object')
        self.assertEqual('42', result)

    def test_from_import_not_existing_method(self):
        path = self.get_contract_path('FromImportNotExistingMethod.py')
        self.assertCompilerLogs(CompilerError.UnresolvedReference, path)

    def test_import_not_existing_method(self):
        path = self.get_contract_path('ImportNotExistingMethod.py')
        self.assertCompilerLogs(CompilerError.UnresolvedReference, path)

    def test_import_boa_invalid_package(self):
        path = self.get_contract_path('ImportBoaInvalidPackage.py')
        self.assertCompilerLogs(CompilerError.UnresolvedReference, path)
